from types import TracebackType
from typing import Callable, Union

from anthropic import MessageStreamEvent, MessageStreamManager
from anthropic.lib.streaming import MessageStream

from ...scribe import scribe


class TextStreamWrapper:
    def __init__(self, wrapper, callback):
        self._wrapper = wrapper
        self._callback = callback

    def __iter__(self):
        return self

    def __next__(self):
        stream = self._wrapper.__enter__()
        event = stream.__next__()
        self._callback(event)
        if event.type == "text":
            if event.text is not None:
                return event.text
        return ""


class StreamWrapper:
    def __init__(
        self,
        mgr: MessageStreamManager,
        callback: Callable[[MessageStreamEvent], None],
    ) -> None:
        self.__mgr = mgr
        self.__callback = callback

    def __enter__(self) -> MessageStream:
        stream = self.__mgr.__enter__()

        # Create a wrapper that processes chunks while still yielding them
        class StreamWithCallback(MessageStream):
            def __init__(self, stream, callback):
                self._stream = stream
                self._callback = callback

            def __iter__(self):
                return self

            def __next__(self):
                chunk = next(self._stream)
                try:
                    self._callback(chunk)
                except Exception as e:
                    scribe().error(f"Error in callback: {e}")
                return chunk

            @property
            def text_stream(self) -> TextStreamWrapper:  # type: ignore
                # Return a wrapper around the original text_stream that calls the callback
                return TextStreamWrapper(self, self._callback)

        return StreamWithCallback(stream, self.__callback)

    def __exit__(
        self,
        exc_type: Union[type[BaseException], None],
        exc: Union[BaseException, None],
        exc_tb: Union[TracebackType, None],
    ) -> None:
        self.__mgr.__exit__(exc_type, exc, exc_tb)



import logging
from typing import Any, Dict, Optional, TypedDict

from ..writer import LogWriter
from .base import BaseContainer, Entity

logger = logging.getLogger("MaximSDK")


class ErrorConfig(TypedDict, total=False):
    id: Optional[str]
    message: str
    name: Optional[str]
    code: Optional[str]
    type: Optional[str]
    tags: Optional[Dict[str, str]]
    metadata: Optional[Dict[str, Any]]


class Error(BaseContainer):
    def __init__(self, config: ErrorConfig, writer: LogWriter):
        final_config = dict(config)
        super().__init__(Entity.ERROR, final_config, writer)
        self.message = final_config.get("message", None)
        self.code = final_config.get("code", None)
        self.error_type = final_config.get("type", None)
        self.tags = final_config.get("tags", None)
        self.metadata = final_config.get("metadata", None)

    def data(self) -> Dict[str, Any]:
        base_data = super().data()
        return {
            **base_data,
            "message": self.message,
            "code": self.code,
            "errorType": self.error_type,
            "metadata": self.metadata,
        }



import time
from collections import OrderedDict
from typing import Any


class ExpiringKeyValueStore:
    def __init__(self):
        self.store = OrderedDict()

    def set(self, key: str, value: Any, expiry_seconds: int):
        expiry_time = time.time() + expiry_seconds
        self.store[key] = (value, expiry_time)
        self._evict_expired()

    def get(self, key: str):
        if key in self.store:
            value, expiry_time = self.store[key]
            if time.time() < expiry_time:
                return value
            else:
                del self.store[key]
        return None

    def delete(self, key: str):
        if key in self.store:
            del self.store[key]

    def _evict_expired(self):
        current_time = time.time()
        expired_keys = [
            key
            for key, (_, expiry_time) in self.store.items()
            if current_time >= expiry_time
        ]
        for key in expired_keys:
            del self.store[key]



import time
from typing import Any, Dict, Iterable, Iterator, List, Optional, Union
from uuid import uuid4

from botocore.client import BaseClient

from ...scribe import scribe
from ..logger import Generation, GenerationConfig, Logger, Trace, TraceConfig
from .async_client import MaximBedrockAsyncClient
from .utils import BedrockUtils

# Define a type alias for message structure for clarity
BedrockMessageParam = Dict[str, Union[str, List[Dict[str, str]]]]
BedrockStreamEvent = Dict[str, Any]
BedrockMessage = Dict[str, Any]

class MaximBedrockClient:
    def __init__(
        self,
        logger: Logger,
        client: BaseClient,
    ):
        self._client = client
        self._logger = logger
        self._async_client = MaximBedrockAsyncClient(logger, client)


    def converse_stream(
        self,
        messages: Iterable[BedrockMessageParam],
        system: Optional[Union[str, List[Dict[str, str]]]] = None,
        *,
        modelId: str,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = None,
        generation_name: Optional[str] = None,
        inferenceConfig: Optional[Dict[str, Any]] = None,
        toolConfig: Optional[Dict[str, Any]] = None,
        guardrailConfig: Optional[Dict[str, Any]] = None,
        performanceConfig: Optional[Dict[str, Any]] = None,
        additionalModelRequestFields: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Iterator[BedrockStreamEvent]:
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None
        model = BedrockUtils.get_model_name(modelId)

        final_inference_config = {
            **(inferenceConfig or {}),
            **({ 'maxTokens': max_tokens } if max_tokens is not None else {}),
            **kwargs.get('inference_config', {})
        }
        final_inference_config = {k: v for k, v in final_inference_config.items() if v is not None}    
        try:
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            generation_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                provider="bedrock",
                name=generation_name,
                model_parameters={
                    "model_id": modelId,
                    **final_inference_config,
                    **(additionalModelRequestFields or {}),
                },
                messages=BedrockUtils.parse_message_param(messages),
            )
            generation = trace.generation(generation_config)
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][BedrockClient] Error preparing logging: {str(e)}"
            )

        api_params = {
            "modelId": modelId,
            "messages": list(messages),
        }
        if system:
            api_params["system"] = [{"text": system}] if isinstance(system, str) else system
        if final_inference_config:
            api_params["inferenceConfig"] = final_inference_config
        if toolConfig:
            api_params["toolConfig"] = toolConfig
        if guardrailConfig:
            api_params["guardrailConfig"] = guardrailConfig
        if performanceConfig:
            api_params["performanceConfig"] = performanceConfig
        if additionalModelRequestFields:
             api_params["additionalModelRequestFields"] = additionalModelRequestFields

        response_stream = self._client.converse_stream(**api_params)
        
        try:
            if generation is not None:
                with response_stream as stream:
                    # Initialize combined response object
                    combined_response = {
                        "id": str(uuid4()),
                        "created": int(time.time()),
                        "choices": [{
                            "index": 0,
                            "message": {
                                "role": "assistant",
                                "content": ""
                            },
                            "finish_reason": None
                        }],
                        "usage": {
                            "prompt_tokens": 0,
                            "completion_tokens": 0,
                            "total_tokens": 0
                        }
                    }
                    
                    for event in stream:
                        event_type = event.get("event-type")
                        
                        if event_type == "contentBlockDelta":
                            delta = event.get("delta", {})
                            if "text" in delta:
                                # Append to the combined content
                                combined_response["choices"][0]["message"]["content"] += delta["text"]
                                
                        elif event_type == "messageStop":
                            # Set finish reason
                            combined_response["choices"][0]["finish_reason"] = event.get("stopReason", "stop")
                            
                        elif event_type == "metadata":
                            usage = event.get("usage", {})
                            if usage:
                                # Update usage information
                                combined_response["usage"] = {
                                    "prompt_tokens": usage.get("inputTokens", 0),
                                    "completion_tokens": usage.get("outputTokens", 0),
                                    "total_tokens": usage.get("totalTokens", 0)
                                }
                            else:
                                scribe().warning(
                                    "[MaximSDK][BedrockClient] No usage metrics available in metadata"
                                )
                        
                        # Pass through the event for streaming
                        yield event
                    
                    # After collecting all events, send the combined response to generation
                    if generation is not None:
                        generation.result(combined_response)
                        if is_local_trace and trace is not None:
                            trace.set_output(combined_response["choices"][0]["message"]["content"])
                            trace.end()

        except Exception as e:
            scribe().warning(
                f"[MaximSDK][BedrockClient] Error processing Bedrock stream or logging: {str(e)}"
            )


    def converse(
        self,
        messages: Iterable[BedrockMessageParam],
        system: Optional[Union[str, List[Dict[str, str]]]] = None,
        *,
        modelId: str,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = None,
        generation_name: Optional[str] = None,
        inferenceConfig: Optional[Dict[str, Any]] = None,
        toolConfig: Optional[Dict[str, Any]] = None,
        guardrailConfig: Optional[Dict[str, Any]] = None,
        performanceConfig: Optional[Dict[str, Any]] = None,
        additionalModelRequestFields: Optional[Dict[str, Any]] = None,
         **kwargs: Any,
    ) -> BedrockMessage:
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None
        model = BedrockUtils.get_model_name(modelId)
        final_inference_config = {
            **(inferenceConfig or {}),
            **({ 'maxTokens': max_tokens } if max_tokens is not None else {}),
            **kwargs.get('inference_config', {})
        }
        final_inference_config = {k: v for k, v in final_inference_config.items() if v is not None}
        try:
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            generation_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                provider="bedrock",
                name=generation_name,
                model_parameters={
                    "model_id": modelId,
                    **final_inference_config,
                     **(additionalModelRequestFields or {}),
                },
                messages=BedrockUtils.parse_message_param(messages),
            )
            generation = trace.generation(generation_config)
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][BedrockClient] Error preparing logging: {str(e)}"
            )
        api_params = {
            "modelId": modelId,
            "messages": list(messages),
        }
        if system:
            api_params["system"] = [{"text": system}] if isinstance(system, str) else system
        if final_inference_config:
            api_params["inferenceConfig"] = final_inference_config
        if toolConfig:
            api_params["toolConfig"] = toolConfig
        if guardrailConfig:
            api_params["guardrailConfig"] = guardrailConfig
        if performanceConfig:
            api_params["performanceConfig"] = performanceConfig
        if additionalModelRequestFields:
             api_params["additionalModelRequestFields"] = additionalModelRequestFields        
        response = self._client.converse(**api_params)
        changed_response = BedrockUtils.parse_message(response)        
        try:
            if generation is not None:
                generation.result(changed_response)
                if is_local_trace and trace is not None:
                    if changed_response is not None:
                        content = changed_response.get("choices", [])[0].get("message", {}).get("content", "")
                        trace.set_output(content)
                    generation.end()
                    trace.end()

        except Exception as e:
            scribe().warning(
                f"[MaximSDK][BedrockClient] Error processing Bedrock stream or logging: {str(e)}"
            )
        return response
                    
    @property
    def async_client(self) -> MaximBedrockAsyncClient:
        return self._async_client



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, TypedDict, Union

from typing_extensions import deprecated

from ..components import FileAttachment, FileDataAttachment, UrlAttachment
from ..writer import LogWriter
from .base import BaseContainer
from .types import Entity


@deprecated(
    "This class will be removed in a future version. Use {} which is TypedDict."
)
@dataclass
class RetrievalConfig():
    id: str
    name: Optional[str] = None
    tags: Optional[Dict[str, str]] = None


class RetrievalConfigDict(TypedDict, total=False):
    id: str
    name: Optional[str]
    tags: Optional[Dict[str, str]]


def get_retrieval_config_dict(
    config: Union[RetrievalConfig, RetrievalConfigDict],
) -> dict[str, Any]:
    return (
        dict(
            RetrievalConfigDict(
                id=config.id,
                name=config.name,
                tags=config.tags,
            )
        )
        if isinstance(config, RetrievalConfig)
        else dict(config)
    )


class Retrieval(BaseContainer):
    def __init__(
        self, config: Union[RetrievalConfig, RetrievalConfigDict], writer: LogWriter
    ):
        final_config = get_retrieval_config_dict(config)
        super().__init__(Entity.RETRIEVAL, dict(final_config), writer)
        self.is_output_set = False

    def input(self, query: str):
        if query is None:
            return
        self._commit("update", {"input": query})
        self.end()

    @staticmethod
    def input_(writer: LogWriter, id: str, query: str):
        BaseContainer._commit_(writer, Entity.RETRIEVAL,
                               id, "update", {"input": query})

    def output(self, docs: Union[str, List[str]]):
        final_docs = docs if isinstance(docs, list) else [docs]
        self.is_output_set = True
        self._commit(
            "update", {"docs": final_docs, "endTimestamp": datetime.now(timezone.utc)})
        self.end()

    def add_attachment(
        self, attachment: Union[FileAttachment, FileDataAttachment, UrlAttachment]
    ):
        self._commit("upload-attachment", attachment.to_dict())

    @staticmethod
    def add_attachment_(
        writer: LogWriter,
        id: str,
        attachment: Union[FileAttachment, FileDataAttachment, UrlAttachment],
    ):
        BaseContainer._commit_(
            writer, Entity.RETRIEVAL, id, "upload-attachment", attachment.to_dict()
        )

    @staticmethod
    def output_(writer: LogWriter, id: str, docs: Union[str, List[str]]):
        final_docs = docs if isinstance(docs, list) else [docs]
        BaseContainer._commit_(writer, Entity.RETRIEVAL, id, "update", {
                               "docs": final_docs})
        BaseContainer._end_(writer, Entity.RETRIEVAL, id, {
                            "endTimestamp": datetime.now(timezone.utc)})

    @staticmethod
    def end_(writer: LogWriter, id: str, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = {}
        BaseContainer._end_(writer, Entity.RETRIEVAL, id, {
            "endTimestamp": datetime.now(timezone.utc),
            **data,
        })

    @staticmethod
    def add_tag_(writer: LogWriter, id: str, key: str, value: str):
        BaseContainer._add_tag_(writer, Entity.RETRIEVAL, id, key, value)



from .dataset import (
    create_data_structure,
    sanitize_data_structure,
    validate_data_structure,
)

__all__ = [
    "create_data_structure",
    "sanitize_data_structure",
    "validate_data_structure",
]



from openai import AsyncOpenAI

from ..logger import (
    Logger,
)
from .async_chat import MaximAsyncOpenAIChat


class MaximOpenAIAsyncClient:
    def __init__(self, client: AsyncOpenAI, logger: Logger):
        self._client = client
        self._logger = logger

    @property
    def chat(self) -> MaximAsyncOpenAIChat:
        return MaximAsyncOpenAIChat(self._client, self._logger)



from datetime import datetime, timezone
from uuid import uuid4

from .store import (
    SessionStoreEntry,
    Turn,
    get_livekit_callback,
    get_maxim_logger,
    get_session_store,
)


def start_new_turn(session_info: SessionStoreEntry):
    """
    This function will start a new turn and return the current turn.
    If the current turn is interrupted or empty, it will return None.
    If the current turn is not interrupted and not empty, it will return the current turn.
    If the current turn is interrupted and not empty, it will return None.
    If the current turn is empty, it will return None.
    If the current turn is not interrupted and empty, it will return None.
    If the current turn is interrupted and empty, it will return None.

    Args:
        session_info: The session information.

    Returns:
        The new turn or None if the current turn is interrupted or empty.
    """
    turn = session_info.get("current_turn", None)
    if turn is not None:
        if turn["is_interrupted"]:
            # User interrupted before assistant response - continue in current turn
            return None
        if (
            turn["turn_input_transcription"] == ""
            and turn["turn_input_audio_buffer"] == b""
        ):
            return None
    trace = get_session_store().get_current_trace_from_rt_session_id(
        session_info["rt_session_id"]
    )
    if trace is not None:
        trace.end()
        if get_livekit_callback() is not None:
            get_livekit_callback()(
                "maxim.trace.ended", {"trace_id": trace.id, "trace": trace}
            )
    next_turn_sequence = 1
    if turn is not None and "turn_sequence" in turn:
        next_turn_sequence = turn["turn_sequence"] + 1
    # Creating a new turn and new trace
    session_id = session_info["mx_session_id"]
    trace_id = str(uuid4())
    tags = {}
    if session_info["room_id"] is not None:
        tags["room_id"] = session_info["room_id"]
    if session_info["agent_id"] is not None:
        tags["agent_id"] = session_info["agent_id"]
    current_turn = Turn(
        turn_id=str(uuid4()),
        turn_sequence=next_turn_sequence,
        turn_timestamp=datetime.now(timezone.utc),
        turn_input_audio_buffer=bytes(),
        is_interrupted=False,
        turn_input_transcription="",
        turn_output_transcription="",
        turn_output_audio_buffer=bytes(),
    )
    trace = get_maxim_logger().trace(
        {
            "id": trace_id,
            "name": f"Turn {next_turn_sequence}",
            "session_id": session_id,
            "tags": tags,
        }
    )
    session_info["user_speaking"] = True
    session_info["current_turn"] = current_turn
    session_info["mx_current_trace_id"] = trace_id
    get_session_store().set_session(session_info)
    if get_livekit_callback() is not None:
        get_livekit_callback()(
            "maxim.trace.started", {"trace_id": trace_id, "trace": trace}
        )
    return current_turn



import time
import weakref
from datetime import datetime
from enum import Enum
from threading import Lock
from typing import Callable, Optional, TypedDict, Union

from livekit.agents import AgentSession
from livekit.agents.llm import RealtimeSession

from ...scribe import scribe
from ..components import FileDataAttachment
from ..logger import Logger, Trace
from ..utils import pcm16_to_wav_bytes


class SessionState(Enum):
    INITIALIZED = 0
    GREETING = 1
    STARTED = 2


class Turn(TypedDict):
    turn_id: str
    turn_sequence: int
    turn_timestamp: datetime
    is_interrupted: bool
    turn_input_transcription: str
    turn_output_transcription: str
    turn_input_audio_buffer: bytes
    turn_output_audio_buffer: bytes


class LLMConfig(TypedDict):
    id: str
    object: str
    expires_at: int
    input_audio_noise_reduction: Optional[bool]
    turn_detection: dict
    input_audio_format: str
    input_audio_transcription: Optional[dict]
    client_secret: Optional[str]
    include: Optional[list]
    model: str
    modalities: list[str]
    instructions: str
    voice: str
    output_audio_format: str
    tool_choice: str
    temperature: float
    max_response_output_tokens: Union[str, int]
    speed: float
    tools: list


class SessionStoreEntry(TypedDict):
    room_id: str
    state: SessionState
    provider: str
    user_speaking: bool
    conversation_buffer_index: int
    conversation_buffer: bytes
    llm_config: Optional[LLMConfig]
    agent_id: Optional[int]
    agent_session_id: Optional[int]
    agent_session: Optional[weakref.ref[AgentSession]]
    rt_session_id: Optional[int]
    rt_session: Optional[weakref.ref[RealtimeSession]]
    mx_current_trace_id: Optional[str]
    mx_session_id: Optional[str]
    rt_session_info: Optional[dict]
    current_turn: Optional[Turn]


MaximLiveKitCallback = Callable[[str, dict], None]
livekit_callback: MaximLiveKitCallback = None

maxim_logger: Union[Logger, None] = None


def set_livekit_callback(callback: MaximLiveKitCallback) -> None:
    global livekit_callback
    livekit_callback = callback


def get_livekit_callback() -> Optional[MaximLiveKitCallback]:
    global livekit_callback
    return livekit_callback


def get_maxim_logger() -> Logger:
    """Get the global maxim logger instance."""
    if maxim_logger is None:
        raise ValueError("Maxim logger is not set")
    return maxim_logger


def set_maxim_logger(logger: Logger) -> None:
    """Set the global maxim logger instance."""
    global maxim_logger
    maxim_logger = logger


class LiveKitSessionStore:
    def __init__(self):
        self.mx_live_kit_session_store: list[SessionStoreEntry] = []
        self._lock = Lock()

    def get_session_by_room_id(self, room_id: str) -> Union[SessionStoreEntry, None]:
        with self._lock:
            for entry in self.mx_live_kit_session_store:
                if "room_id" in entry and entry["room_id"] == room_id:
                    return entry
            return None

    def get_session_by_agent_session_id(
        self, session_id: int
    ) -> Union[SessionStoreEntry, None]:
        with self._lock:
            for entry in self.mx_live_kit_session_store:
                if (
                    "agent_session_id" in entry
                    and entry["agent_session_id"] == session_id
                ):
                    return entry
            return None

    def get_session_by_rt_session_id(
        self, rt_session_id: int
    ) -> Union[SessionStoreEntry, None]:
        with self._lock:
            for entry in self.mx_live_kit_session_store:
                if "rt_session_id" in entry and entry["rt_session_id"] == rt_session_id:
                    return entry
            return None

    def set_session(self, entry: SessionStoreEntry):
        with self._lock:
            if "agent_session_id" in entry:
                # find the entry and replace
                for i, e in enumerate(self.mx_live_kit_session_store):
                    if (
                        "agent_session_id" in e
                        and e["agent_session_id"] == entry["agent_session_id"]
                    ):
                        self.mx_live_kit_session_store[i] = entry
                        return
            self.mx_live_kit_session_store.append(entry)

    def delete_session(self, agent_session_id: int):
        with self._lock:
            # Use list comprehension to avoid modifying list while iterating
            self.mx_live_kit_session_store = [
                entry
                for entry in self.mx_live_kit_session_store
                if not (
                    "agent_session_id" in entry
                    and entry["agent_session_id"] == agent_session_id
                )
            ]

    def clear_all_sessions(self):
        with self._lock:
            self.mx_live_kit_session_store.clear()

    def get_current_trace_for_agent_session(
        self, agent_session_id: int
    ) -> Union[Trace, None]:
        session = self.get_session_by_agent_session_id(agent_session_id)
        if session is None:
            return None
        trace_id = session["mx_current_trace_id"]
        if trace_id is None:
            return None
        return get_maxim_logger().trace({"id": trace_id})

    def get_current_trace_for_room_id(self, room_id: str) -> Union[Trace, None]:
        session = self.get_session_by_room_id(room_id)
        if session is None:
            return None
        trace_id = session["mx_current_trace_id"]
        if trace_id is None:
            return None
        return get_maxim_logger().trace({"id": trace_id})

    def get_current_trace_from_rt_session_id(
        self, rt_session_id: int
    ) -> Union[Trace, None]:
        session = self.get_session_by_rt_session_id(rt_session_id)
        if session is None:
            return None
        trace_id = session["mx_current_trace_id"]
        if trace_id is None:
            return None
        return get_maxim_logger().trace({"id": trace_id})

    def get_all_sessions(self):
        with self._lock:
            return self.mx_live_kit_session_store.copy()

    def close_session(self, session_info: SessionStoreEntry):
        with self._lock:
            scribe().debug(
                f"[MaximSDK] Closing session {session_info['mx_session_id']}"
            )
            session_id = session_info["mx_session_id"]
            index = session_info["conversation_buffer_index"]

            get_maxim_logger().session_add_attachment(
                session_id,
                FileDataAttachment(
                    data=pcm16_to_wav_bytes(session_info["conversation_buffer"]),
                    tags={"attach-to": "input"},
                    name=f"Conversation part {index}",
                    timestamp=int(time.time()),
                ),
            )
            get_maxim_logger().session_end(session_id=session_id)


# Create a thread-local storage for the session store
_session_store = LiveKitSessionStore()


def get_session_store():
    """Get the global session store instance."""
    return _session_store



from dataclasses import dataclass
from typing import Dict, List, Union

from maxim.evaluators.base_evaluator import BaseEvaluator

from ..models.evaluator import (
    Evaluator,
    EvaluatorType,
    PassFailCriteria,
)
from ..utils.utils import create_cuid_generator


@dataclass
class EvaluatorNameToIdAndPassFailCriteria:
    id: str
    pass_fail_criteria: PassFailCriteria

    def __init__(self, id: str, pass_fail_criteria: PassFailCriteria):
        self.id = id
        self.pass_fail_criteria = pass_fail_criteria


def get_local_evaluator_name_to_id_and_pass_fail_criteria_map(
    evaluators: List[Union[BaseEvaluator, str]],
) -> Dict[str, EvaluatorNameToIdAndPassFailCriteria]:
    all_eval_names: List[str] = []
    for evaluator in evaluators:
        if isinstance(evaluator, str):
            pass
        else:
            all_eval_names.extend(evaluator.names)

    all_pass_fail_criteria: Dict[str, PassFailCriteria] = {}
    for evaluator in evaluators:
        if isinstance(evaluator, BaseEvaluator):
            all_pass_fail_criteria.update(evaluator.pass_fail_criteria)

    name_to_id_and_pass_fail_criteria_map: Dict[
        str, EvaluatorNameToIdAndPassFailCriteria
    ] = {}
    for eval_name in all_eval_names:
        generate_cuid = create_cuid_generator()
        name_to_id_and_pass_fail_criteria_map[eval_name] = (
            EvaluatorNameToIdAndPassFailCriteria(
                id=generate_cuid(),
                pass_fail_criteria=all_pass_fail_criteria[eval_name],
            )
        )

    return name_to_id_and_pass_fail_criteria_map


def get_evaluator_config_from_evaluator_name_and_pass_fail_criteria(
    id: str, name: str, pass_fail_criteria: PassFailCriteria
) -> Evaluator:
    return Evaluator(
        builtin=False,
        id=id,
        name=name,
        reversed=False,
        type=EvaluatorType.LOCAL,
        config={
            "passFailCriteria": {
                "entryLevel": {
                    "value": (
                        pass_fail_criteria.on_each_entry.value
                        if not (
                            isinstance(pass_fail_criteria.on_each_entry.value, bool)
                        )
                        else (
                            "Yes"
                            if pass_fail_criteria.on_each_entry.value is True
                            else "No"
                        )
                    ),
                    "operator": pass_fail_criteria.on_each_entry.score_should_be,
                    "name": "score",
                },
                "runLevel": {
                    "value": pass_fail_criteria.for_testrun_overall.value,
                    "operator": pass_fail_criteria.for_testrun_overall.overall_should_be,
                    "name": (
                        "meanScore"
                        if pass_fail_criteria.for_testrun_overall.for_result
                        == "average"
                        else "queriesPassed"
                    ),
                },
            },
        },
    )



from .utils import AnthropicUtils
from .client import MaximAnthropicClient
from .async_client import MaximAnthropicAsyncClient

__all__ = ["AnthropicUtils","MaximAnthropicAsyncClient","MaximAnthropicClient"]


import asyncio
from typing import Awaitable, List, Optional, Dict

from maxim.evaluators.base_evaluator import BaseEvaluator

from ..models.dataset import DataStructure, LocalData, Variable
from ..models.evaluator import (
    LocalEvaluationResult,
    LocalEvaluatorResultParameter,
    LocalEvaluatorReturn,
)


async def process_awaitable(awaitable: Awaitable):
    return await awaitable

def get_variables_from_row(
    row: LocalData, data_structure: DataStructure
) -> Dict[str, Variable]:
    variables = {}
    for column_name, column_type in data_structure.items():
        if column_type == "FILE_URL_VARIABLE":
            variables[column_name] = Variable(
                type="file", payload={"files": [{"url": row[column_name]}]}
            )
        elif column_type in ("VARIABLE", "NULLABLE_VARIABLE"):
            # Skip nullable variables with None values to avoid invalid payloads
            if column_type == "NULLABLE_VARIABLE" and row[column_name] is None:
                continue
            variables[column_name] = Variable(
                type="text", payload={"text": row[column_name]}
            )
    return variables


def get_input_expected_output_and_context_from_row(
    input_key: Optional[str],
    expectedOutputKey: Optional[str],
    contextToEvaluateKey: Optional[str],
    row: LocalData,
):
    input = None
    expected_output = None
    context_to_evaluate = None
    if input_key is not None and input_key in row:
        input = str(row[input_key]) if row[input_key] is not None else None
    if expectedOutputKey is not None and expectedOutputKey in row:
        expected_output = (
            str(row[expectedOutputKey]) if row[expectedOutputKey] is not None else None
        )
    if contextToEvaluateKey is not None and contextToEvaluateKey in row:
        context_to_evaluate = (
            str(row[contextToEvaluateKey])
            if row[contextToEvaluateKey] is not None
            else None
        )
    return input, expected_output, context_to_evaluate


async def run_local_evaluations(
    evaluators: List[BaseEvaluator],
    data_entry: LocalData,
    processed_data: LocalEvaluatorResultParameter,
) -> List[LocalEvaluationResult]:
    coroutines = [
        asyncio.to_thread(evaluator.guarded_evaluate, processed_data, data_entry)
        for evaluator in evaluators
    ]
    evaluator_results = await asyncio.gather(*coroutines)
    results: List[LocalEvaluationResult] = []
    for i, evaluator in enumerate(evaluators):
        if isinstance(evaluator, BaseEvaluator):
            try:
                combined_results = evaluator_results[i]
                for name, result in combined_results.items():
                    results.append(
                        LocalEvaluationResult(
                            name=name,
                            pass_fail_criteria=evaluator.pass_fail_criteria[name],
                            result=result,
                        )
                    )
            except Exception as err:
                results.extend(
                    [
                        LocalEvaluationResult(
                            name=name,
                            pass_fail_criteria=evaluator.pass_fail_criteria[name],
                            result=LocalEvaluatorReturn(
                                score="Err",
                                reasoning=f"Error while running combined evaluator with names {evaluator.names}: {str(err)}",
                            ),
                        )
                        for name in evaluator.names
                    ]
                )
    return results



import time
from typing import Any, AsyncGenerator, Dict, Iterable, List, Optional, Union
from uuid import uuid4

from botocore.client import BaseClient

from ...scribe import scribe
from ..logger import Generation, GenerationConfig, Logger, Trace, TraceConfig
from .utils import BedrockUtils

# Define a type alias for message structure for clarity
BedrockMessageParam = Dict[str, Union[str, List[Dict[str, str]]]]
BedrockStreamEvent = Dict[str, Any]
BedrockMessage = Dict[str, Any]

class MaximBedrockAsyncClient:
    def __init__(
        self,
        logger: Logger,
        client: BaseClient,        
    ):
        self._client = client
        self._logger = logger

    async def converse_stream(
        self,
        messages: Iterable[BedrockMessageParam],
        system: Optional[Union[str, List[Dict[str, str]]]] = None,
        *,
        model: str,
        modelId: str,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = None,
        generation_name: Optional[str] = None,
        inferenceConfig: Optional[Dict[str, Any]] = None,
        toolConfig: Optional[Dict[str, Any]] = None,
        guardrailConfig: Optional[Dict[str, Any]] = None,
        performanceConfig: Optional[Dict[str, Any]] = None,
        additionalModelRequestFields: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> AsyncGenerator[BedrockStreamEvent, None]:
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None

        final_inference_config = {
            **(inferenceConfig or {}),
            **({ 'maxTokens': max_tokens } if max_tokens is not None else {}),
            **kwargs.get('inference_config', {})
        }
        final_inference_config = {k: v for k, v in final_inference_config.items() if v is not None}

        try:
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            generation_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                provider="bedrock",
                name=generation_name,
                model_parameters={
                    "model_id": modelId,
                    **final_inference_config,
                    **(additionalModelRequestFields or {}),
                },
                messages=BedrockUtils.parse_message_param(messages),
            )
            generation = trace.generation(generation_config)
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][BedrockAsyncClient] Error preparing logging: {str(e)}"
            )

        api_params = {
            "modelId": modelId,
            "messages": list(messages),
        }
        if system:
            api_params["system"] = [{"text": system}] if isinstance(system, str) else system
        if final_inference_config:
            api_params["inferenceConfig"] = final_inference_config
        if toolConfig:
            api_params["toolConfig"] = toolConfig
        if guardrailConfig:
            api_params["guardrailConfig"] = guardrailConfig
        if performanceConfig:
            api_params["performanceConfig"] = performanceConfig
        if additionalModelRequestFields:
             api_params["additionalModelRequestFields"] = additionalModelRequestFields

        response_stream = await self._client.converse_stream(**api_params)
        
        try:
            if generation is not None:
                # Initialize combined response object (similar to sync version)
                combined_response = {
                    "id": str(uuid4()),
                    "created": int(time.time()),
                    "choices": [{
                        "index": 0,
                        "message": {
                            "role": "assistant",
                            "content": ""
                        },
                        "finish_reason": None
                    }],
                    "usage": {
                        "prompt_tokens": 0,
                        "completion_tokens": 0,
                        "total_tokens": 0
                    }
                }

                async with response_stream as stream:
                    async for event in stream:
                        event_type = event.get("event-type")
                        
                        if event_type == "contentBlockDelta":
                            delta = event.get("delta", {})
                            if "text" in delta:
                                # Append to the combined content
                                combined_response["choices"][0]["message"]["content"] += delta["text"]
                                
                        elif event_type == "messageStop":
                            # Set finish reason
                            combined_response["choices"][0]["finish_reason"] = event.get("stopReason", "stop")
                            
                        elif event_type == "metadata":
                            usage = event.get("usage", {})
                            if usage:
                                # Update usage information
                                combined_response["usage"] = {
                                    "prompt_tokens": usage.get("inputTokens", 0),
                                    "completion_tokens": usage.get("outputTokens", 0),
                                    "total_tokens": usage.get("totalTokens", 0)
                                }
                            else:
                                scribe().warning(
                                    "[MaximSDK][BedrockAsyncClient] No usage metrics available in metadata"
                                )
                            
                        # Pass through the event for streaming
                        yield event
                
                # After collecting all events, send the combined response to generation
                if generation is not None:
                    generation.result(combined_response) 
                    if is_local_trace and trace is not None:
                        trace.set_output(combined_response["choices"][0]["message"]["content"])
                        trace.end()

        except Exception as e:
            scribe().warning(
                f"[MaximSDK][BedrockAsyncClient] Error processing Bedrock stream or logging: {str(e)}"
            )

    async def converse(
        self,
        messages: Iterable[BedrockMessageParam],
        system: Optional[Union[str, List[Dict[str, str]]]] = None,
        *,
        model: str,
        modelId: str,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = None,
        generation_name: Optional[str] = None,
        inferenceConfig: Optional[Dict[str, Any]] = None,
        toolConfig: Optional[Dict[str, Any]] = None,
        guardrailConfig: Optional[Dict[str, Any]] = None,
        performanceConfig: Optional[Dict[str, Any]] = None,
        additionalModelRequestFields: Optional[Dict[str, Any]] = None,
         **kwargs: Any,
    ) -> BedrockMessage:
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None

        final_inference_config = {
            **(inferenceConfig or {}),
            **({ 'maxTokens': max_tokens } if max_tokens is not None else {}),
            **kwargs.get('inference_config', {})
        }
        final_inference_config = {k: v for k, v in final_inference_config.items() if v is not None}
        
        try:
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            generation_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                provider="bedrock",
                name=generation_name,
                model_parameters={
                    "model_id": modelId,
                    **final_inference_config,
                     **(additionalModelRequestFields or {}),
                },
                messages=BedrockUtils.parse_message_param(messages),
            )
            generation = trace.generation(generation_config)
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][BedrockAsyncClient] Error preparing logging: {str(e)}"
            )

        api_params = {
            "modelId": modelId,
            "messages": list(messages),
        }
        if system:
            api_params["system"] = [{"text": system}] if isinstance(system, str) else system
        if final_inference_config:
            api_params["inferenceConfig"] = final_inference_config
        if toolConfig:
            api_params["toolConfig"] = toolConfig
        if guardrailConfig:
            api_params["guardrailConfig"] = guardrailConfig
        if performanceConfig:
            api_params["performanceConfig"] = performanceConfig
        if additionalModelRequestFields:
             api_params["additionalModelRequestFields"] = additionalModelRequestFields
        
        response = await self._client.converse(**api_params)
        changed_response = BedrockUtils.parse_message(response)

        try:
            if generation is not None:
                generation.result(changed_response)
                if is_local_trace and trace is not None:
                    if changed_response is not None:
                        content = changed_response.get("choices", [])[0].get("message", {}).get("content", "")
                        trace.set_output(content)
                    generation.end()
                    trace.end()

        except Exception as e:
            scribe().warning(
                f"[MaximSDK][BedrockAsyncClient] Error processing Bedrock stream or logging: {str(e)}"
            )

        return response
                    


"""Portkey AI integration for Maxim logging.

This module provides one-line integration with Portkey AI clients,
enabling automatic logging of OpenAI-compatible calls via Maxim.
"""

from .client import instrument_portkey
from .portkey import (
    MaximPortkeyClient,
)

__all__ = [
    "MaximPortkeyClient",
    "instrument_portkey",
]



import time
from typing import Any, Dict, Iterable, List, Optional, Union

from openai._streaming import Stream
from openai.types.chat import ChatCompletion, ChatCompletionChunk

from ..logger import GenerationRequestMessage


class OpenAIUtils:
    @staticmethod
    def parse_message_param(
        messages: Iterable[Dict[str, Any]],
        override_role: Optional[str] = None,
    ) -> List[GenerationRequestMessage]:
        parsed_messages: List[GenerationRequestMessage] = []

        for msg in messages:
            role = override_role or msg.get("role", "user")
            content = msg.get("content", "")

            if isinstance(content, list):
                # Handle content blocks for multimodal
                text_content = ""
                for block in content:
                    if isinstance(block, dict) and block.get("type") == "text":
                        text_content += block.get("text", "")
                parsed_messages.append(
                    GenerationRequestMessage(role=role, content=text_content)
                )
            else:
                parsed_messages.append(
                    GenerationRequestMessage(role=role, content=str(content))
                )

        return parsed_messages

    @staticmethod
    def get_model_params(
        **kwargs: Any,
    ) -> Dict[str, Any]:
        model_params = {}
        skip_keys = ["messages"]
        max_tokens = kwargs.get("max_tokens", None)
        if max_tokens is not None:
            model_params["max_tokens"] = max_tokens

        param_keys = [
            "temperature",
            "top_p",
            "presence_penalty",
            "frequency_penalty",
            "response_format",
            "tools",
            "tool_choice",
        ]
        for key in param_keys:
            if key in kwargs and kwargs[key] is not None and key not in skip_keys:
                model_params[key] = kwargs[key]

        for key, value in kwargs.items():
            if key not in param_keys and key not in skip_keys and value is not None:
                model_params[key] = value

        return model_params

    @staticmethod
    def parse_stream_response(
        chunk: ChatCompletionChunk,
    ) -> Dict[str, Any]:
        return {
            "id": chunk.id,
            "created": int(time.time()),
            "choices": [
                {
                    "index": choice.index,
                    "delta": {
                        "role": "assistant",
                        "content": choice.delta.content or "",
                        "tool_calls": getattr(choice.delta, "tool_calls", None),
                    },
                    "finish_reason": choice.finish_reason,
                }
                for choice in chunk.choices
            ],
        }

    @staticmethod
    def parse_completion(
        completion: Union[ChatCompletion, Stream[ChatCompletionChunk]],
    ) -> Dict[str, Any]:
        if isinstance(completion, Stream):
            # Process the stream of chunks
            chunks = []
            for chunk in completion:
                chunks.append(OpenAIUtils.parse_stream_response(chunk))
            # Combine all chunks into a single response
            if chunks:
                last_chunk = chunks[-1]
                combined_content = "".join(
                    [
                        choice.get("delta", {}).get("content", "")
                        for chunk in chunks
                        for choice in chunk.get("choices", [])
                    ]
                )
                return {
                    "id": last_chunk.get("id", ""),
                    "created": int(time.time()),
                    "choices": [
                        {
                            "index": choice.get("index", 0),
                            "message": {
                                "role": "assistant",
                                "content": combined_content,
                                "tool_calls": choice.get("delta", {}).get("tool_calls"),
                            },
                            "finish_reason": choice.get("finish_reason"),
                        }
                        for choice in last_chunk.get("choices", [])
                    ],
                }
            else:
                return {}

        return {
            "id": completion.id,
            "created": int(time.time()),
            "choices": [
                {
                    "index": choice.index,
                    "message": {
                        "role": "assistant",
                        "content": choice.message.content,
                        "tool_calls": [
                            tool_call.model_dump()
                            for tool_call in (choice.message.tool_calls or [])
                        ],
                    },
                    "finish_reason": choice.finish_reason,
                }
                for choice in completion.choices
            ],
            "usage": {
                "prompt_tokens": completion.usage.prompt_tokens
                if completion.usage
                else 0,
                "completion_tokens": completion.usage.completion_tokens
                if completion.usage
                else 0,
                "total_tokens": completion.usage.total_tokens
                if completion.usage
                else 0,
            },
        }



import contextvars
import functools
import importlib
import inspect
import logging
import uuid
from time import time
from typing import Union

from crewai import LLM, Agent, Crew, Flow, Task
from crewai.agents.agent_builder.base_agent import BaseAgent
from crewai.tools.base_tool import BaseTool

from ...logger import (
    Generation,
    GenerationConfigDict,
    Logger,
    Retrieval,
    Span,
    SpanConfigDict,
    ToolCall,
    Trace,
)
from ...scribe import scribe
from .utils import (
    crew_kickoff_postprocess_inputs,
    crewai_postprocess_inputs,
    dictify,
    extract_tool_details,
    extract_tool_name,
    get_agent_display_name,
    get_task_display_name,
)

_last_llm_usages = {}
_task_span_ids = {}

_global_maxim_trace: contextvars.ContextVar[Union[Trace, None]] = (
    contextvars.ContextVar("maxim_trace_context_var", default=None)
)


def get_log_level(debug: bool) -> int:
    """
    Set logging level based on debug flag.
    debug=False: Only WARNING and ERROR logs
    debug=True: INFO and DEBUG logs
    """
    return logging.DEBUG if debug else logging.WARNING


class MaximUsageCallback:
    def __init__(self, generation_id: str):
        self.generation_id = generation_id

    def log_success_event(self, kwargs, response_obj, start_time, end_time):
        global _last_llm_usages
        usage_info = response_obj.get("usage")
        if usage_info:
            if isinstance(usage_info, dict):
                _last_llm_usages[self.generation_id] = usage_info
            elif hasattr(usage_info, "prompt_tokens"):
                _last_llm_usages[self.generation_id] = {
                    "prompt_tokens": getattr(usage_info, "prompt_tokens", 0),
                    "completion_tokens": getattr(usage_info, "completion_tokens", 0),
                    "total_tokens": getattr(usage_info, "total_tokens", 0),
                }
            else:
                _last_llm_usages[self.generation_id] = None  # Couldn't parse
            scribe().debug(
                f"[MaximSDK] Callback captured usage: {_last_llm_usages[self.generation_id] is not None}"
            )
        else:
            _last_llm_usages[self.generation_id] = None
            scribe().debug(
                "[MaximSDK] Callback did not find usage info in response_obj"
            )


# --- Wrapper Factory for _handle_non_streaming_response ---
def make_handle_non_streaming_wrapper(original_method):
    @functools.wraps(original_method)
    def handle_non_streaming_wrapper(
        self,
        params: dict,
        callbacks: Union[list, None] = None,
        available_functions: Union[dict, None] = None,
    ):
        _maxim_generation_id = getattr(self, "_maxim_generation_id", None)
        if not isinstance(_maxim_generation_id, str):
            scribe().warning(
                "[MaximSDK] No generation ID found for LLM call. Skipping usage capture."
            )
            return original_method(self, params, callbacks, available_functions)

        custom_callback = MaximUsageCallback(_maxim_generation_id)

        # Ensure callbacks is a list and add our custom one
        current_callbacks = callbacks if callbacks is not None else []
        if not isinstance(current_callbacks, list):  # Safety check
            scribe().warning(
                "[MaximSDK] Original callbacks is not a list, creating new list."
            )
            current_callbacks = []
        current_callbacks.append(custom_callback)

        # Call the original method with the augmented callbacks
        result = original_method(self, params, current_callbacks, available_functions)
        return result

    return handle_non_streaming_wrapper


def instrument_crewai(maxim_logger: Logger, debug: bool = False):
    """
    Patches CrewAI's core components (Crew, Agent, Task, Flow, LLM) to add comprehensive logging and tracing.

    This wrapper enhances CrewAI with:
    - Detailed operation tracing for Crew, Flow, and Task executions
    - Token usage tracking for LLM calls
    - Tool execution monitoring
    - Span-based operation tracking
    - Error handling and reporting

    The patching is done by wrapping key methods like:
    - Crew.kickoff
    - Agent.execute_task
    - Task.execute_sync
    - LLM.call and _handle_non_streaming_response
    - Tool._run methods

    Args:
        maxim_logger (Logger): A Maxim Logger instance for handling the tracing and logging operations.
        debug (bool): If True, show INFO and DEBUG logs. If False, show only WARNING and ERROR logs.
    """
    global logger

    # Set the logging level based on debug parameter
    scribe().setLevel(get_log_level(debug))

    def make_maxim_wrapper(
        original_method,
        base_op_name: str,
        input_processor=None,
        output_processor=None,
        display_name_fn=None,
    ):
        @functools.wraps(original_method)
        def maxim_wrapper(self, *args, **kwargs):
            scribe().debug(f"――― Start: {base_op_name} ―――")

            global _global_maxim_trace
            global _task_span_ids

            # Combine args and kwargs into a dictionary for processing
            bound_args = {}
            processed_inputs = {}
            final_op_name = base_op_name

            try:
                sig = inspect.signature(original_method)
                bound_values = sig.bind(self, *args, **kwargs)
                bound_values.apply_defaults()
                bound_args = bound_values.arguments

                # Process inputs if processor is provided
                processed_inputs = bound_args
                if input_processor:
                    try:
                        processed_inputs = input_processor(bound_args)
                    except Exception as e:
                        scribe().debug(
                            f"[MaximSDK] Failed to process inputs for {base_op_name}: {e}"
                        )

                if display_name_fn:
                    try:
                        final_op_name = display_name_fn(processed_inputs)
                    except Exception as e:
                        scribe().debug(
                            f"[MaximSDK] Failed to generate display name for {base_op_name}: {e}"
                        )

            except Exception as e:
                scribe().debug(
                    f"[MaximSDK] Failed to bind/process inputs for {base_op_name}: {e}"
                )
                # Fallback for inputs display
                processed_inputs = {"self": self, "args": args, "kwargs": kwargs}

            trace: Union[Trace, None] = None
            span: Union[Span, None] = None
            generation: Union[Generation, None] = None
            tool_call: Union[ToolCall, Retrieval, None] = None
            planner_span: Union[Span, None] = None
            trace_token: Union[contextvars.Token[Union[Trace, None]], None] = None

            if isinstance(self, Flow):
                if _global_maxim_trace.get() is None:
                    trace_id = str(uuid.uuid4())
                    scribe().debug(f"Creating trace for flow [{trace_id}]")

                    trace = maxim_logger.trace(
                        {
                            "id": trace_id,
                            "name": "Flow Kickoff",
                            "tags": {
                                "flow_id": str(getattr(self, "id", "")),
                                "flow_name": final_op_name,
                            },
                            "input": str(processed_inputs["inputs"] or "-"),
                        }
                    )

                    # Store the token for later restoration
                    trace_token = _global_maxim_trace.set(trace)

            elif isinstance(self, Crew):
                if _global_maxim_trace.get() is None:
                    trace_id = str(uuid.uuid4())

                    if original_method.__name__ == "kickoff_for_each":
                        scribe().debug(
                            "[MaximSDK] Creating trace for crew kickoff_for_each"
                        )

                        trace = maxim_logger.trace(
                            {
                                "id": trace_id,
                                "name": "Crew Kickoff For Each",
                                "tags": {
                                    "crew_id": str(getattr(self, "id", "")),
                                },
                                "input": str(processed_inputs["inputs"] or "-"),
                            }
                        )

                    else:
                        scribe().debug(
                            f"[MaximSDK] Creating trace for crew [{trace_id}]"
                        )

                        trace = maxim_logger.trace(
                            {
                                "id": trace_id,
                                "name": "Crew Kickoff",
                                "tags": {
                                    "crew_id": str(getattr(self, "id", "")),
                                    "crew_name": final_op_name,
                                },
                                "input": str(processed_inputs["inputs"] or "-"),
                            }
                        )

                        # Attach trace to all tasks in the crew
                        if hasattr(self, "tasks") and self.tasks:
                            scribe().debug(
                                f"[MaximSDK] Attaching trace to {len(self.tasks)} tasks"
                            )
                            for task in self.tasks:
                                setattr(task, "_trace", trace)
                                scribe().debug(
                                    f"[MaximSDK] Task: {task.description[:40]}{'...' if len(task.description) > 40 else ''}"
                                )

                    trace_token = _global_maxim_trace.set(trace)
                else:
                    span_id = str(uuid.uuid4())

                    if original_method.__name__ == "kickoff_for_each":
                        scribe().debug(
                            f"[MaximSDK] Attaching span to crew kickoff_for_each [{span_id}]"
                        )

                        span = _global_maxim_trace.get().span(
                            {
                                "id": span_id,
                                "name": "Crew Kickoff For Each",
                                "tags": {
                                    "crew_id": str(getattr(self, "id", "")),
                                },
                            }
                        )
                    else:
                        scribe().debug(f"[MaximSDK] Attaching span to crew [{span_id}]")

                        span = _global_maxim_trace.get().span(
                            {
                                "id": span_id,
                                "name": "Crew Kickoff",
                                "tags": {
                                    "crew_id": str(getattr(self, "id", "")),
                                    "crew_name": final_op_name,
                                },
                            }
                        )

                        # Attach trace to all tasks in the crew
                        if hasattr(self, "tasks") and self.tasks:
                            scribe().debug(
                                f"[MaximSDK] Attaching trace to {len(self.tasks)} tasks"
                            )
                            for task in self.tasks:
                                setattr(task, "_span", span)
                                scribe().debug(
                                    f"[MaximSDK] Task: {task.description[:40]}{'...' if len(task.description) > 40 else ''}"
                                )

            elif isinstance(self, Task):
                # Get the trace from the crew
                trace = getattr(self, "_trace", None)
                if not isinstance(trace, Trace):
                    trace = None

                span = getattr(self, "_span", None)
                if not isinstance(span, Span):
                    span = None

                span_id = str(uuid.uuid4())

                scribe().debug(
                    f"[MaximSDK] Task span [{span_id}] for '{self.name or self.description[:40]}'"
                )

                task_span_config = SpanConfigDict(
                    {
                        "id": span_id,
                        "name": f"Task: {self.name or 'None'}",
                        "tags": {"task_id": str(getattr(self, "id", ""))},
                    }
                )

                if not trace and not span:  # Will happen for Planner tasks
                    # TODO get crew info for flow workflows crew placement
                    scribe().debug(
                        "[MaximSDK] Parent trace/span not found, creating new task span on global trace"
                    )
                    if _global_maxim_trace.get() is None:
                        scribe().warning(
                            "[MaximSDK] No global trace found, skipping logging"
                        )
                        return
                    span = _global_maxim_trace.get().span(task_span_config)
                else:
                    if trace:
                        scribe().debug(f"[MaximSDK] Found parent trace: {trace.id}")
                        span = trace.span(task_span_config)
                    else:
                        if not span:
                            scribe().warning(
                                "[MaximSDK] No parent span found, skipping logging"
                            )
                            return
                        scribe().debug(f"[MaximSDK] Found parent span: {span.id}")
                        span = span.span(task_span_config)

                metadata = {
                    "name": self.name or "None",
                    "description": self.description or "None",
                    "expected_output": self.expected_output or "None",
                }

                if self.output_file:
                    metadata["output_file"] = self.output_file

                if self.output_json:
                    metadata["output_json"] = self.output_json

                if self.max_retries:
                    metadata["max_retries"] = self.max_retries

                if self.tools:
                    metadata["tools"] = [tool.name for tool in self.tools]

                span.add_metadata(metadata)

                _task_span_ids[self.id] = span_id

                if self.agent is not None:
                    scribe().debug(
                        f"[MaximSDK] Attaching span to agent [{self.agent.id}]"
                    )
                    setattr(self.agent, "_span", span)
                else:
                    # Check if agent is provided in args/kwargs
                    agent_from_kwargs = kwargs.get("agent", None) if kwargs else None
                    agent_from_args = None
                    args_list = list(args)  # Convert to list to allow modification

                    if not agent_from_kwargs and args:
                        # If first arg is agent, use it (common pattern in execute_sync/execute_core)
                        if len(args) > 0 and isinstance(args[0], BaseAgent):
                            agent_from_args = args[0]

                    # Use whichever agent we found
                    agent_to_use = agent_from_kwargs or agent_from_args

                    if agent_to_use:
                        scribe().debug(
                            f"[MaximSDK] Found agent in args, attaching span to agent [{agent_to_use.role}]"
                        )
                        setattr(agent_to_use, "_span", span)

                        # Update the agent in its original location
                        if agent_from_kwargs:
                            kwargs["agent"] = agent_to_use
                        elif agent_from_args:
                            args_list[0] = agent_to_use
                            args = tuple(args_list)  # Convert back to tuple
                    else:
                        scribe().warning("[MaximSDK] Task has no agent assigned")

                trace = None

            elif isinstance(self, Agent):
                span = getattr(self, "_span", None)
                if not isinstance(span, Span):
                    span = None

                span_id = str(uuid.uuid4())

                agent_span_config = SpanConfigDict(
                    {
                        "id": span_id,
                        "name": f"Agent: {self.role}",
                        "tags": {"agent_id": str(getattr(self, "id", ""))},
                    }
                )

                skip_logging = False

                if span:
                    scribe().debug(
                        f"[MaximSDK] Agent span [{span_id}] for '{self.role}'"
                    )
                    span = span.span(agent_span_config)
                else:
                    scribe().debug("[MaximSDK] Agent has no span, checking task")

                    # First check args/kwargs for task
                    task: Union[Task, None] = None
                    if len(args) > 0:
                        for arg in args:
                            if isinstance(arg, Task):
                                task = arg
                                break

                    if not task and kwargs:  # Check kwargs if task not found in args
                        for kwarg_value in kwargs.values():
                            if isinstance(kwarg_value, Task):
                                task = kwarg_value
                                break

                    # Fallback to agent_executor.task if no task found in args or kwargs
                    if not task and hasattr(self, "agent_executor"):
                        task = self.agent_executor.task

                    if task:
                        span_id = _task_span_ids.get(task.id)
                        if span_id:
                            span = maxim_logger.span_add_sub_span(
                                span_id, agent_span_config
                            )
                        else:
                            scribe().debug(
                                f"[MaximSDK] No span found for task {task.id}, creating new task span"
                            )
                            # Create a new task span since none exists
                            task_span_id = str(uuid.uuid4())
                            task_config = SpanConfigDict(
                                {
                                    "id": task_span_id,
                                    "name": f"Task: {task.name or task.description}",
                                    "tags": {
                                        "task_id": str(task.id),
                                    },
                                }
                            )
                            if (
                                _global_maxim_trace.get() is not None
                            ):  # TODO check for check in flows
                                span = _global_maxim_trace.get().span(task_config)
                                _task_span_ids[task.id] = task_span_id
                                # Now create the agent span as a child of the task span
                                span = span.span(agent_span_config)
                            else:
                                scribe().debug(
                                    "[MaximSDK] No global trace found, skipping logging"
                                )
                                skip_logging = True
                    else:
                        scribe().warning(
                            f"[MaximSDK] Agent {self.role} has no task or span, skipping logging"
                        )
                        skip_logging = True

                if not skip_logging:
                    if hasattr(self, "llm") and self.llm:
                        scribe().debug(
                            f"[MaximSDK] LLM: {getattr(self.llm, 'model', 'unknown')}"
                        )
                    if isinstance(self.llm, LLM):
                        setattr(self.llm, "_span", span)

                    # Check for tools in both agent's tools attribute and in execute_task arguments
                    tools_to_set_context = []

                    # Add tools from agent's tools attribute
                    if hasattr(self, "tools") and self.tools:
                        # Filter for BaseTool instances from agent's tools
                        for tool in self.tools:
                            if (
                                isinstance(tool, BaseTool)
                                and tool not in tools_to_set_context
                            ):
                                tools_to_set_context.append(tool)

                    # Check for tools in execute_task arguments
                    if original_method.__name__ == "execute_task":
                        # Check kwargs for tools
                        tools_from_kwargs = (
                            kwargs.get("tools", None) if kwargs else None
                        )
                        if tools_from_kwargs:
                            if isinstance(tools_from_kwargs, list):
                                # Filter for BaseTool instances from kwargs list
                                for tool in tools_from_kwargs:
                                    if (
                                        isinstance(tool, BaseTool)
                                        and tool not in tools_to_set_context
                                    ):
                                        tools_to_set_context.append(tool)
                            elif (
                                isinstance(tools_from_kwargs, BaseTool)
                                and tools_from_kwargs not in tools_to_set_context
                            ):
                                tools_to_set_context.append(tools_from_kwargs)

                        # Check args for tools (usually after task argument)
                        if len(args) > 1 and isinstance(args[1], (list, BaseTool)):
                            tools_from_args = args[1]
                            if isinstance(tools_from_args, list):
                                # Filter for BaseTool instances from args list
                                for tool in tools_from_args:
                                    if (
                                        isinstance(tool, BaseTool)
                                        and tool not in tools_to_set_context
                                    ):
                                        tools_to_set_context.append(tool)
                            elif (
                                isinstance(tools_from_args, BaseTool)
                                and tools_from_args not in tools_to_set_context
                            ):
                                tools_to_set_context.append(tools_from_args)

                    # Set context for all collected tools
                    if tools_to_set_context:
                        scribe().debug(
                            f"[MaximSDK] Attaching span to {len(tools_to_set_context)} unique valid tools"
                        )
                        for tool in tools_to_set_context:
                            setattr(tool, "_span", span)

            elif isinstance(self, LLM):
                span = getattr(self, "_span", None)
                if not isinstance(span, Span):
                    span = None

                generation_id = str(uuid.uuid4())
                setattr(self, "_maxim_generation_id", generation_id)
                scribe().debug(f"[MaximSDK] LLM generation [{generation_id}]")

                llm_generation_config = GenerationConfigDict(
                    {
                        "id": generation_id,
                        "name": "LLM Call",
                        "provider": (
                            "anthropic" if self.is_anthropic else "openai"
                        ),  # TODO: Add more providers
                        "model": str(getattr(self, "model", "unknown")),
                        "messages": args[0],
                    }
                )

                if span:
                    llm_generation_config["span_id"] = span.id
                    generation = span.generation(llm_generation_config)

                    span = None
                else:
                    scribe().warning(
                        "[MaximSDK] No parent span found for LLM call, creating new planner span"
                    )
                    if _global_maxim_trace.get() is None:
                        scribe().warning(
                            "[MaximSDK] No global trace found, skipping logging"
                        )
                        return
                    planner_span = _global_maxim_trace.get().span(
                        {
                            "id": str(uuid.uuid4()),
                            "name": "Planner",
                        }
                    )
                    llm_generation_config["span_id"] = planner_span.id

                    generation = planner_span.generation(llm_generation_config)

            elif isinstance(self, BaseTool):
                span = getattr(self, "_span", None)
                if not isinstance(span, Span):
                    span = None

                if span:
                    tool_id = str(uuid.uuid4())
                    tool_name = extract_tool_name(final_op_name)

                    if tool_name == "RagTool":
                        scribe().debug(f"[MaximSDK] RAG: Retrieval tool [{tool_id}]")
                        tool_call = span.retrieval(
                            {
                                "id": tool_id,
                                "name": f"RAG: {self.name}",
                                "tags": {"span_id": span.id},
                            }
                        )
                        tool_call.input(processed_inputs["query"])
                    else:
                        scribe().debug(f"[MaximSDK] TOOL: {self.name} [{tool_id}]")

                        tool_details = extract_tool_details(self.description)
                        tool_call = span.tool_call(
                            {
                                "id": tool_id,
                                "name": f"{tool_details['name'] or self.name}",
                                "description": tool_details["description"]
                                or self.description,
                                "args": (
                                    str(tool_details["args"])
                                    if tool_details["args"] is not None
                                    else str(processed_inputs)
                                ),
                                "tags": {"tool_id": tool_id, "span_id": span.id},
                            }
                        )

                    span = None
                else:
                    scribe().warning("[MaximSDK] No parent span found for tool call")

            scribe().debug(f"\n[MaximSDK] --- Calling: {final_op_name} ---")

            try:
                # Call the original method (bound to self)
                output = original_method.__get__(self, self.__class__)(*args, **kwargs)
            except Exception as e:
                scribe().error(f"[MaximSDK] {type(e).__name__} in {final_op_name}")

                if tool_call:
                    if isinstance(tool_call, Retrieval):
                        tool_call.output(f"Error occurred while calling tool: {e}")
                        scribe().debug("[MaximSDK] RAG: Completed retrieval with error")
                    else:
                        tool_call.result(f"Error occurred while calling tool: {e}")
                        scribe().debug(
                            "[MaximSDK] TOOL: Completed tool call with error"
                        )

                if generation:
                    generation.error({"message": str(e)})
                    scribe().debug("[MaximSDK] GEN: Completed generation with error")

                if span:
                    span.add_error({"message": str(e)})
                    span.end()
                    scribe().debug("[MaximSDK] SPAN: Completed span with error")

                if trace and "crewai.Crew" in final_op_name:
                    trace.add_error({"message": str(e)})
                    trace.end()
                    scribe().debug("[MaximSDK] TRACE: Completed trace with error")

                    # Get the stored token and reset the context
                    if trace_token is not None:
                        _global_maxim_trace.reset(trace_token)
                    else:
                        _global_maxim_trace.set(None)

                    maxim_logger.flush()

                raise e  # Re-raise the original exception

            processed_output = output
            if output_processor:
                try:
                    processed_output = output_processor(output)
                except Exception as e:
                    scribe().debug(f"[MaximSDK] Failed to process output: {e}")

            if tool_call:
                if isinstance(tool_call, Retrieval):
                    tool_call.output(processed_output)
                    scribe().debug("[MaximSDK] RAG: Completed retrieval")
                else:
                    tool_call.result(processed_output)
                    scribe().debug("[MaximSDK] TOOL: Completed tool call")

            if generation:
                # Create a structured result compatible with GenerationResult
                # Retrieve usage data captured by the callback
                global _last_llm_usages  # Ensure access to the global

                prompt_tokens = 0
                completion_tokens = 0
                total_tokens = 0

                usage_data = _last_llm_usages.get(generation.id)

                if usage_data and isinstance(usage_data, dict):
                    prompt_tokens = usage_data.get("prompt_tokens", 0)
                    completion_tokens = usage_data.get("completion_tokens", 0)
                    total_tokens = usage_data.get("total_tokens", 0)
                    scribe().debug(
                        f"[MaximSDK] GEN: Using captured token usage: P={prompt_tokens}, C={completion_tokens}, T={total_tokens}"
                    )
                else:
                    scribe().debug(
                        f"[MaximSDK] GEN: Using default token usage (0). Captured data: {usage_data}"
                    )

                result = {
                    "id": f"gen_{generation.id}",
                    "object": "chat.completion",
                    "created": int(time()),
                    "model": str(getattr(self, "model", "unknown")),
                    "choices": [
                        {
                            "index": 0,
                            "message": {
                                "role": "assistant",
                                "content": str(processed_output),
                            },
                            "finish_reason": "stop",
                        }
                    ],
                    "usage": {
                        "prompt_tokens": prompt_tokens,
                        "completion_tokens": completion_tokens,
                        "total_tokens": total_tokens,
                    },
                }

                generation.result(result)
                scribe().debug("[MaximSDK] GEN: Completed generation")

                if planner_span:
                    planner_span.end()
                    scribe().debug("[MaximSDK] PLANNER: Completed planner")

                del _last_llm_usages[generation.id]

            if span:
                scribe().debug("[MaximSDK] SPAN: Completing span")
                span.end()

            if trace:
                scribe().debug(f"[MaximSDK] TRACE: Completing trace [{trace.id}]")
                if processed_output is not None:
                    trace.set_output(str(processed_output))
                trace.end()

                # Get the stored token and reset the context
                if trace_token is not None:
                    _global_maxim_trace.reset(trace_token)
                else:
                    _global_maxim_trace.set(None)

                maxim_logger.flush()

            scribe().debug(f"――― End: {final_op_name} ―――\n")

            return output

        return maxim_wrapper

    # --- 0. Patch LLM._handle_non_streaming_response to capture usage ---
    if LLM is not None and hasattr(LLM, "_handle_non_streaming_response"):
        original_handle_method = getattr(LLM, "_handle_non_streaming_response")
        wrapper_handle = make_handle_non_streaming_wrapper(original_handle_method)
        setattr(LLM, "_handle_non_streaming_response", wrapper_handle)
        scribe().info(
            "[MaximSDK] Patched crewai.LLM._handle_non_streaming_response to capture usage."
        )

    # --- 1. Patch Crew Methods ---
    crew_methods_to_patch = ["kickoff"]
    for method_name in crew_methods_to_patch:
        if hasattr(Crew, method_name):
            original_method = getattr(Crew, method_name)
            op_name = f"crewai.Crew.{method_name}"
            wrapper = make_maxim_wrapper(
                original_method,
                op_name,
                input_processor=crew_kickoff_postprocess_inputs,
            )
            setattr(Crew, method_name, wrapper)
            scribe().info(f"[MaximSDK] Patched crewai.Crew.{method_name} for printing.")

    # --- 2. Patch Agent.execute_task ---
    agent_methods_to_patch = ["execute_task"]
    for method_name in agent_methods_to_patch:
        if hasattr(Agent, method_name):
            original_method = getattr(Agent, method_name)
            op_name = f"crewai.Crew.{method_name}"
            wrapper = make_maxim_wrapper(
                original_method,
                op_name,
                input_processor=crewai_postprocess_inputs,
                display_name_fn=get_agent_display_name,
            )
            setattr(Agent, method_name, wrapper)
            scribe().info(
                f"[MaximSDK] Patched crewai.Agent.{method_name} for printing."
            )

    # --- 3. Patch Task.execute_sync ---
    task_methods_to_patch = ["execute_sync", "execute_async"]
    for method_name in task_methods_to_patch:
        if hasattr(Task, method_name):
            original_method = getattr(Task, method_name)
            op_name = f"crewai.Task.{method_name}"
            wrapper = make_maxim_wrapper(
                original_method,
                op_name,
                input_processor=crewai_postprocess_inputs,
                display_name_fn=get_task_display_name,
            )
            setattr(Task, method_name, wrapper)
            scribe().info(f"[MaximSDK] Patched crewai.Task.{method_name} for printing.")

    # --- 4. Patch LLM.call ---
    if LLM is not None and hasattr(LLM, "call"):
        original_method = getattr(LLM, "call")
        op_name = "crewai.LLM.call"

        wrapper = make_maxim_wrapper(
            original_method,
            op_name,
            input_processor=lambda inputs: dictify(inputs),
            output_processor=lambda output: dictify(output),
        )
        setattr(LLM, "call", wrapper)
        scribe().info("[MaximSDK] Patched crewai.LLM.call for printing.")

    # --- 5. Patch CrewAI Tools ---
    try:
        crewai_tools_module = importlib.import_module("crewai_tools")
        tool_names = [
            t
            for t in dir(crewai_tools_module)
            if "Tool" in t and not t.startswith("Base")
        ]

        for tool_name in tool_names:
            try:
                tool_class = getattr(crewai_tools_module, tool_name)
                if (
                    isinstance(tool_class, type)
                    and issubclass(tool_class, BaseTool)
                    and hasattr(tool_class, "_run")
                ):
                    original_tool_run = getattr(tool_class, "_run")
                    op_name = f"crewai_tools.{tool_name}._run"
                    wrapper = make_maxim_wrapper(
                        original_tool_run,
                        op_name,
                        input_processor=lambda inputs: dictify(inputs),
                        output_processor=lambda output: dictify(output),
                    )
                    setattr(tool_class, "_run", wrapper)
                    scribe().info(f"[MaximSDK] Patched {op_name} for printing.")
            except (AttributeError, TypeError, ImportError) as e:
                scribe().warning(
                    f"[MaximSDK] Skipping patching for tool {tool_name}: {e}"
                )
                continue
    except ImportError:
        scribe().warning(
            "[MaximSDK] crewai_tools or BaseTool not found. Skipping tool patching."
        )
    except Exception as e:
        scribe().error(f"[MaximSDK] ERROR during tool patching: {e}")

    # --- 6. Patch Flow Methods ---
    if Flow is not None:
        # Patch Flow kickoff methods
        flow_kickoff_methods = ["kickoff", "kickoff_async"]
        for method_name in flow_kickoff_methods:
            if hasattr(Flow, method_name):
                original_method = getattr(Flow, method_name)
                op_name = f"crewai.Flow.{method_name}"
                wrapper = make_maxim_wrapper(
                    original_method,
                    op_name,
                    input_processor=lambda inputs: dictify(inputs),
                    output_processor=lambda output: dictify(output),
                )
                setattr(Flow, method_name, wrapper)
                scribe().info(
                    f"[MaximSDK] Patched crewai.Flow.{method_name} for printing."
                )

    scribe().info("[MaximSDK] Finished applying patches to CrewAI.")



from .langchain import langchain_callback, langchain_llm_call, langgraph_agent

try:
    import langchain
except ImportError:
    raise ImportError(
        "Langchain is not installed. Please install it to use these decorators."
    )


__all__ = [
    "langchain_callback",
    "langchain_llm_call",
    "langgraph_agent",
]



from openai import OpenAI
from openai.resources.chat import Chat

from ..logger import (
    Logger,
)
from .completions import MaximOpenAIChatCompletions


class MaximOpenAIChat(Chat):
    def __init__(self, client: OpenAI, logger: Logger):
        super().__init__(client=client)
        self._logger = logger

    @property
    def completions(self) -> MaximOpenAIChatCompletions:
        return MaximOpenAIChatCompletions(self._client, self._logger)



import asyncio
from contextvars import ContextVar
from functools import wraps
from typing import Callable, Dict, List, Optional, Union
from uuid import uuid4

from ..logger.components.trace import Trace, TraceConfig
from ..logger.logger import Logger

_maxim_ctx_var_trace: ContextVar[Optional[Trace]] = ContextVar(
    "maxim_ctx_trace", default=None
)

_maxim_ctx_var_logger: ContextVar[Optional[Logger]] = ContextVar(
    "maxim_ctx_logger", default=None
)


def current_logger() -> Optional[Logger]:
    try:
        return _maxim_ctx_var_logger.get()
    except LookupError:
        return None


def current_trace() -> Optional[Trace]:
    try:
        return _maxim_ctx_var_trace.get()
    except LookupError:
        return None


def trace(
    logger: Logger,
    id: Optional[Union[str, Callable]] = None,
    sessionId: Optional[Union[str, Callable]] = None,
    name: Optional[str] = None,
    tags: Optional[dict] = None,
    evaluators: Optional[List[str]] = None,
    evaluator_variables: Optional[Dict[str, str]] = None,
):
    def decorator(func):
        if asyncio.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                actual_trace_id = id() if callable(id) else id
                if actual_trace_id is None:
                    actual_trace_id = str(uuid4())
                trace_config = TraceConfig(id=actual_trace_id, name=name, tags=tags)
                # If session_id is present, we will attach that to the trace
                actual_session_id = sessionId() if callable(sessionId) else sessionId
                if actual_session_id is not None:
                    trace_config.session_id = actual_session_id
                trace = logger.trace(config=trace_config)
                if evaluators is not None:
                    trace.evaluate().with_evaluators(*evaluators).with_variables(
                        evaluator_variables or {}
                    )
                token = _maxim_ctx_var_trace.set(trace)
                _maxim_ctx_var_logger.set(logger)
                try:
                    result = await func(*args, **kwargs)
                    if trace.data().get("output", None) is None:
                        trace.set_output(str(result))
                    return result
                finally:
                    trace.end()
                    _maxim_ctx_var_trace.reset(token)

            return async_wrapper
        else:

            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                actual_trace_id = id() if callable(id) else id
                if actual_trace_id is None:
                    actual_trace_id = str(uuid4())
                trace_config = TraceConfig(id=actual_trace_id, name=name, tags=tags)
                # If session_id is present, we will attach that to the trace
                actual_session_id = sessionId() if callable(sessionId) else sessionId
                if actual_session_id is not None:
                    trace_config.session_id = actual_session_id
                trace = logger.trace(config=trace_config)
                if evaluators is not None:
                    trace.evaluate().with_evaluators(*evaluators).with_variables(
                        evaluator_variables or {}
                    )
                token = _maxim_ctx_var_trace.set(trace)
                _maxim_ctx_var_logger.set(logger)
                try:
                    result = func(*args, **kwargs)
                    if trace.data().get("output", None) is None:
                        trace.set_output(str(result))
                    return result
                finally:
                    trace.end()
                    _maxim_ctx_var_trace.reset(token)

            return sync_wrapper

    return decorator



import asyncio
import logging
from contextvars import ContextVar
from functools import wraps
from typing import Callable, Dict, List, Optional, Union
from uuid import uuid4

from ..logger.components.generation import Generation, GenerationConfig
from ..logger.logger import Logger
from .span import current_span
from .trace import current_logger, current_trace

_generation_ctx_var: ContextVar[Optional[Generation]] = ContextVar(
    "maxim_ctx_generation", default=None
)


def current_generation() -> Optional[Generation]:
    return _generation_ctx_var.get()


def generation(
    logger: Optional[Logger] = None,
    id: Optional[Union[Callable,str]] = None,
    name: Optional[str] = None,
    maxim_prompt_id: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    evaluators: Optional[List[str]] = None,
    evaluator_variables: Optional[Dict[str, str]] = None,
):
    def decorator(func):
        if asyncio.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger
                if maxim_logger is None:
                    maxim_logger = current_logger()
                if maxim_logger is None:
                    raise ValueError(
                        "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                    )
                if current_span() is None and current_trace() is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                actual_generation_id = id() if callable(id) else id
                if actual_generation_id is None:
                    actual_generation_id = str(uuid4())
                generation_config = GenerationConfig(
                    id=actual_generation_id,
                    name=name,
                    model="unknown",
                    provider="unknown",
                    maxim_prompt_id=maxim_prompt_id,
                    tags=tags,
                )
                generation: Optional[Generation] = None
                if (span := current_span()) is not None:
                    generation = span.generation(config=generation_config)
                elif (trace := current_trace()) is not None:
                    generation = trace.generation(config=generation_config)
                if generation is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    return
                if evaluators is not None:
                    generation.evaluate().with_evaluators(*evaluators).with_variables(
                        evaluator_variables if evaluator_variables is not None else {}
                    )
                token = _generation_ctx_var.set(generation)
                try:
                    return await func(*args, **kwargs)
                finally:
                    generation.end()
                    _generation_ctx_var.reset(token)

        else:
            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger or current_logger()
                if maxim_logger is None:
                    raise ValueError(
                        "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                    )

                if current_span() is None and current_trace() is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                actual_generation_id = id() if callable(id) else id
                if actual_generation_id is None:
                    actual_generation_id = str(uuid4())
                generation_config = GenerationConfig(
                    id=actual_generation_id,
                    name=name,
                    model="unknown",
                    provider="unknown",
                    maxim_prompt_id=maxim_prompt_id,
                    tags=tags,
                )
                generation: Optional[Generation] = None
                if (span := current_span()) is not None:
                    generation = span.generation(config=generation_config)
                elif (trace := current_trace()) is not None:
                    generation = trace.generation(config=generation_config)
                if generation is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    return
                token = _generation_ctx_var.set(generation)
                try:
                    return func(*args, **kwargs)
                finally:
                    generation.end()
                    _generation_ctx_var.reset(token)
            return sync_wrapper

    return decorator



import time
from typing import Any, Dict, Iterable, List, Optional
from uuid import uuid4

from anthropic.types import MessageParam, MessageStreamEvent, TextDelta

from ..logger import GenerationRequestMessage


class AnthropicUtils:
    @staticmethod
    def parse_message_param(
        message: Iterable[MessageParam],
        override_role: Optional[str] = None,
    ) -> List[GenerationRequestMessage]:
        messages: List[GenerationRequestMessage] = []
        
        for msg in message:
            if isinstance(msg, str):
                messages.append(
                    GenerationRequestMessage(
                        role=override_role or "user",
                        content=msg
                    )
                )
            elif isinstance(msg, dict):
                role = override_role or msg.get("role", "user")
                content = msg.get("content", "")
                if isinstance(content, list):
                    # Handle content blocks
                    text_content = ""
                    for block in content:
                        if isinstance(block, dict) and block.get("type") == "text":
                            text_content += block.get("text", "")
                    messages.append(
                        GenerationRequestMessage(
                            role=role,
                            content=text_content
                        )
                    )
                else:
                    messages.append(
                        GenerationRequestMessage(
                            role=role,
                            content=str(content)
                        )
                    )
        
        return messages

    @staticmethod
    def get_model_params(
        max_tokens: int,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        model_params = {}
        
        # Check max_tokens parameter
        if max_tokens is not None:
            model_params["max_tokens"] = max_tokens
            
        # Common parameters to check from kwargs
        param_keys = ["system", "metadata", "temperature", "top_p", "top_k"]
        for key in param_keys:
            if key in kwargs and kwargs[key] is not None:
                model_params[key] = kwargs[key]
            
        # Add any additional parameters from kwargs
        for key, value in kwargs.items():
            if key not in param_keys and value is not None:
                model_params[key] = value
                
        return model_params

    @staticmethod
    def parse_message_stream(
        stream: List[MessageStreamEvent],
    ) -> Dict[str, Any]:
        if not stream:
            raise ValueError("No response chunks")
            
        text = ""
        usage = {
            "prompt_tokens": 0,
            "completion_tokens": 0,
            "total_tokens": 0
        }
        
        for event in stream:
            if hasattr(event, "type"):
                if event.type == "message_start":
                    usage["prompt_tokens"] = event.message.usage.input_tokens
                    usage["completion_tokens"] = event.message.usage.output_tokens
                    usage["total_tokens"] = event.message.usage.input_tokens + event.message.usage.output_tokens
                elif event.type == "content_block_delta":
                    if isinstance(event.delta, TextDelta):
                        text += event.delta.text
                
        return {
            "id": str(uuid4()),
            "created": int(time.time()),
            "choices": [{
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": text,
                },
                "finish_reason": "stop"  # Anthropic doesn't provide this directly
            }],
            "usage": usage
        }

    @staticmethod
    def parse_message(
        message: Any,
    ) -> Dict[str, Any]:
        """
        Parse the Message response from Anthropic API into a standardized format.
        """
        content = ""
        if isinstance(message.content, list):
            for block in message.content:
                # Check if block is a dict and has the expected structure
                if hasattr(block, "type") and hasattr(block, "text") and block.type == "text":
                        content += block.text
                elif isinstance(block, dict) and block.get("type") == "text":
                    content += block.get("text", "")
                elif isinstance(block, dict) and block.get("type") == "message":
                    content += block.get("text", "")
        else:
            content = str(message.content)
            
        return {
            "id": message.id,
            "created": int(time.time()),
            "choices": [{
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": content,
                },
                "finish_reason": message.stop_reason or "stop"
            }],
            "usage": {
                "prompt_tokens": message.usage.input_tokens,
                "completion_tokens": message.usage.output_tokens,
                "total_tokens": message.usage.input_tokens + message.usage.output_tokens
            }
        }


import asyncio
import json
import logging
from contextvars import ContextVar
from functools import wraps
from typing import Callable, Dict, List, Optional, Union
from uuid import uuid4

from ..logger.components.tool_call import ToolCall, ToolCallConfig
from ..logger.logger import Logger
from .span import current_span
from .trace import current_logger, current_trace

_tool_call_ctx_var: ContextVar[Optional[ToolCall]] = ContextVar(
    "maxim_ctx_tool_call", default=None
)


def current_tool_call() -> Optional[ToolCall]:
    return _tool_call_ctx_var.get()


def tool_call(
    logger: Optional[Logger] = None,
    id: Optional[Union[str,Callable]] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    arguments: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    evaluators: Optional[List[str]] = None,
    evaluator_variables: Optional[Dict[str, str]] = None,
):

    def decorator(func):
        if asyncio.iscoroutinefunction(func):
            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger or current_logger()
                if maxim_logger is None:
                    raise ValueError(
                        "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                    )
                # Here there should be an active span or active trace
                # If none of this is the case then we raise an error
                if current_span() is None and current_trace() is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                actual_tool_call_id = id() if callable(id) else id
                if actual_tool_call_id is None:
                    actual_tool_call_id = str(uuid4())
                config = ToolCallConfig(
                    id=actual_tool_call_id,
                    name=name or "",
                    description=description or "",
                    args=arguments or "",
                    tags=tags or {},
                )
                tool_call: Optional[ToolCall] = None
                if (span := current_span()) is not None:
                    tool_call = span.tool_call(config)
                elif (trace := current_trace()) is not None:
                    tool_call = trace.tool_call(config)
                if tool_call is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    return
                if evaluators is not None:
                    tool_call.evaluate().with_evaluators(*evaluators).with_variables(evaluator_variables if evaluator_variables is not None else {})
                token = _tool_call_ctx_var.set(tool_call)
                try:
                    result = await func(*args, **kwargs)
                    # Here we will check if tool_call has output
                    if isinstance(result, (dict)):
                        tool_call.result(json.dumps(result))
                    elif isinstance(result, list):
                        tool_call.result(
                            json.dumps([json.dumps(item) for item in result])
                        )
                    elif isinstance(result, str):
                        tool_call.result(result)
                    return result
                finally:
                    tool_call.end()
                    _tool_call_ctx_var.reset(token)
            return async_wrapper
        else:

            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger or current_logger()
                if maxim_logger is None:
                        raise ValueError(
                            "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                        )
                # Here there should be an active span or active trace
                # If none of this is the case then we raise an error
                if current_span() is None and current_trace() is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                actual_tool_call_id = id() if callable(id) else id
                if actual_tool_call_id is None:
                    actual_tool_call_id = str(uuid4())
                config = ToolCallConfig(
                    id=actual_tool_call_id,
                    name=name or "",
                    description=description or "",
                    args=arguments or "",
                    tags=tags or {},
                )
                tool_call: Optional[ToolCall] = None
                if (span:=current_span()) is not None:
                    tool_call = span.tool_call(config)
                elif (trace := current_trace()) is not None:
                    tool_call = trace.tool_call(config)
                if tool_call is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    return
                token = _tool_call_ctx_var.set(tool_call)
                try:
                    result = func(*args, **kwargs)
                    # Here we will check if tool_call has output
                    if isinstance(result, (dict)):
                        tool_call.result(json.dumps(result))
                    elif isinstance(result, list):
                        tool_call.result(
                            json.dumps([json.dumps(item) for item in result])
                        )
                    elif isinstance(result, str):
                        tool_call.result(result)
                    return result
                finally:
                    tool_call.end()
                    _tool_call_ctx_var.reset(token)
            return sync_wrapper

    return decorator



from dataclasses import dataclass
from typing import Any, TypedDict, Union

from typing_extensions import deprecated


class FeedbackDict(TypedDict, total=False):
    score: int
    comment: str


@deprecated(
    "This class will be removed in a future version. Use {} which is TypedDict."
)
@dataclass
class Feedback():
    score: int
    comment: str


def get_feedback_dict(feedback: Union[Feedback, FeedbackDict]) -> dict[str, Any]:
    return (
        dict(
            FeedbackDict(
                score=feedback.score,
                comment=feedback.comment,
            )
        )
        if isinstance(feedback, Feedback)
        else dict(feedback)
    )



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, TypedDict, Union

from typing_extensions import deprecated

from ..writer import LogWriter
from .base import BaseContainer
from .types import Entity


@deprecated(
    "This class will be removed in a future version. Use {} which is TypedDict."
)
@dataclass
class ToolCallConfig:
    id: str
    name: str
    description: str
    args: str
    tags: Optional[Dict[str, str]] = None


class ToolCallConfigDict(TypedDict, total=False):
    id: str
    name: str
    description: str
    args: str
    tags: Optional[Dict[str, str]]


def get_tool_call_config_dict(
    config: Union[ToolCallConfig, ToolCallConfigDict],
) -> ToolCallConfigDict:
    return (
        ToolCallConfigDict(
            id=config.id,
            name=config.name,
            description=config.description,
            args=config.args,
            tags=config.tags,
        )
        if isinstance(config, ToolCallConfig)
        else config
    )


@deprecated(
    "This class will be removed in a future version. Use {} which is TypedDict instead."
)
@dataclass
class ToolCallError:
    message: str
    code: Optional[str] = None
    type: Optional[str] = None


class ToolCallErrorDict(TypedDict):
    message: str
    code: Optional[str]
    type: Optional[str]


def get_tool_call_error_dict(
    error: Union[ToolCallError, ToolCallErrorDict],
) -> dict[str, Any]:
    return dict(
        ToolCallErrorDict(
            message=error.message,
            code=error.code,
            type=error.type,
        )
        if isinstance(error, ToolCallError)
        else dict(error)
    )


class ToolCall(BaseContainer):
    def __init__(
        self, config: Union[ToolCallConfig, ToolCallConfigDict], writer: LogWriter
    ):
        final_config = get_tool_call_config_dict(config)
        if "id" not in final_config:
            raise ValueError("ID is required")
        super().__init__(Entity.TOOL_CALL, dict(final_config), writer)
        self._id = final_config.get("id")
        self._name = final_config.get("name", None)
        self.args = final_config.get("args", None)
        self.description = final_config.get("description", None)
        self.tags = final_config.get("tags", None)

    def update(self, data: Dict[str, Any]):
        self._commit("update", data)

    @staticmethod
    def update_(writer: LogWriter, id: str, data: Dict[str, Any]):
        BaseContainer._commit_(writer, Entity.TOOL_CALL, id, "update", data)

    @staticmethod
    def result_(writer: LogWriter, id: str, result: str):
        BaseContainer._commit_(
            writer, Entity.TOOL_CALL, id, "result", {"result": result}
        )
        BaseContainer._end_(
            writer,
            Entity.TOOL_CALL,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
            },
        )

    def attach_evaluators(self, evaluators: List[str]):
        raise NotImplementedError("attach_evaluators is not supported for ToolCall")

    def with_variables(self, for_evaluators: List[str], variables: Dict[str, str]):
        raise NotImplementedError("with_variables is not supported for ToolCall")

    def result(self, result: str):
        self._commit("result", {"result": result})
        self.end()

    def error(self, error: ToolCallError):
        self._commit("error", {"error": error})
        self.end()

    @staticmethod
    def error_(
        writer: LogWriter, id: str, error: Union[ToolCallError, ToolCallErrorDict]
    ):
        final_error = get_tool_call_error_dict(error)
        BaseContainer._commit_(
            writer, Entity.TOOL_CALL, id, "error", {"error": final_error}
        )
        BaseContainer._end_(
            writer,
            Entity.TOOL_CALL,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
            },
        )

    def data(self) -> Dict[str, Any]:
        base_data = super().data()
        return {
            **base_data,
            "name": self._name,
            "description": self.description,
            "args": self.args,
        }



import json
import re
import uuid
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from ...scribe import scribe
from ..parsers.tags_parser import parse_tags
from ..utils import make_object_serializable
from ..writer import LogWriter
from .types import CommitLog, Entity


class ContainerLister:
    def on_end(self):
        pass


BaseConfig = Dict[str, Any]


def _sanitize_metadata(metadata: Dict[str, Any]) -> Dict[str, str]:
    sanitized_metadata: dict[str, str] = {}
    for key, value in metadata.items():
        serialized_obj = make_object_serializable(value)
        if isinstance(serialized_obj, str):
            sanitized_metadata[key] = serialized_obj
            continue
        sanitized_metadata[key] = json.dumps(serialized_obj)
    return sanitized_metadata


class EvaluateContainerWithVariables:
    def __init__(
        self, id: str, entity: Entity, log_writer: LogWriter, for_evaluators: List[str]
    ) -> None:
        self.entity = entity
        self.writer = log_writer
        self.id = id
        self.for_evaluators = for_evaluators

    def with_variables(self, variables: Dict[str, str]):
        if len(self.for_evaluators) == 0:
            return
        self.writer.commit(
            CommitLog(
                self.entity,
                self.id,
                "evaluate",
                {
                    "with": "variables",
                    "variables": variables,
                    "evaluators": list(set(self.for_evaluators)),
                    "timestamp": datetime.now(timezone.utc),
                },
            )
        )


class EvaluateContainer:
    """
    A class to manage evaluators for a specific entity.

    This class provides functionality to initialize and manage a set of evaluators
    associated with a particular entity and writer.

    Attributes:
        entity (Entity): The entity associated with these evaluators.
        writer (LogWriter): The log writer used for committing evaluator actions.
        evaluators (List[str]): A list of evaluator identifiers.
        id (str): A unique identifier for this set of evaluators.

    Methods:
        with_variables: Allows adding variables to be used by the evaluators.
    """

    def __init__(self, id: str, entity: Entity, log_writer: LogWriter) -> None:
        self.entity = entity
        self.writer = log_writer
        self.id = id

    def with_variables(self, variables: Dict[str, str], for_evaluators: List[str]):
        if len(for_evaluators) == 0:
            raise ValueError("At least one evaluator must be provided")

        self.writer.commit(
            CommitLog(
                self.entity,
                self.id,
                "evaluate",
                {
                    "with": "variables",
                    "variables": variables,
                    "evaluators": list(set(for_evaluators)),
                    "timestamp": datetime.now(timezone.utc),
                },
            )
        )

    def with_evaluators(self, *evaluators: str) -> EvaluateContainerWithVariables:
        if len(evaluators) == 0:
            raise ValueError("At least one evaluator must be provided")

        self.writer.commit(
            CommitLog(
                self.entity,
                self.id,
                "evaluate",
                {
                    "with": "evaluators",
                    "evaluators": list(set(evaluators)),
                    "timestamp": datetime.now(timezone.utc),
                },
            )
        )

        return EvaluateContainerWithVariables(
            self.id, self.entity, self.writer, list(set(evaluators))
        )


class BaseContainer:
    def __init__(self, entity: Entity, config: BaseConfig, writer: LogWriter):
        self.entity = entity
        if "id" not in config:
            self._id = str(uuid.uuid4())
        else:
            self._id = config["id"]
        self._name = config.get("name", None)
        self.span_id = config.get("span_id", None)
        self.start_timestamp = datetime.now(timezone.utc)
        self.end_timestamp = None
        self.tags = parse_tags(config.get("tags", {}))
        self.writer = writer
        # Doing it at the end to avoid problems with regular flow
        # We drop these logs at LogWriter level as well
        # Validate ID format - only allow alphanumeric characters, hyphens, and underscores
        if not re.match(r"^[a-zA-Z0-9_-]+$", self._id):
            if writer.raise_exceptions:
                raise ValueError(
                    f"Invalid ID: {self._id}. ID must only contain alphanumeric characters, hyphens, and underscores. Event will not be logged."
                )
            else:
                scribe().error(
                    f"[MaximSDK] Invalid ID: {config['id']}. ID must only contain alphanumeric characters, hyphens, and underscores. Event will not be logged."
                )

    @property
    def id(self) -> str:
        return self._id

    def evaluate(self) -> EvaluateContainer:
        return EvaluateContainer(self._id, self.entity, self.writer)

    @staticmethod
    def _evaluate_(writer: LogWriter, entity: Entity, id: str) -> EvaluateContainer:
        return EvaluateContainer(id, entity, writer)

    def add_metadata(self, metadata: Dict[str, Any]) -> None:
        sanitized_metadata: dict[str, str] = _sanitize_metadata(metadata)
        self._commit("update", {"metadata": sanitized_metadata})

    @staticmethod
    def add_metadata_(
        writer: LogWriter, entity: Entity, id: str, metadata: Dict[str, Any]
    ) -> None:
        sanitized_metadata: dict[str, str] = _sanitize_metadata(metadata)
        writer.commit(CommitLog(entity, id, "update", {"metadata": sanitized_metadata}))

    def add_tag(self, key: str, value: str):
        if self.tags is None:
            self.tags = {}
        if not isinstance(value, str):
            raise ValueError("Tag value must be a string")
        # Validate if value is str and not None
        if not value:
            raise ValueError("Tag value must not be empty")
        self.tags[key] = value
        self.tags = parse_tags(self.tags)
        self._commit("update", {"tags": {key: value}})

    @staticmethod
    def _add_tag_(writer: LogWriter, entity: Entity, id: str, key: str, value: str):
        writer.commit(CommitLog(entity, id, "update", {"tags": {key: value}}))

    def end(self):
        self.end_timestamp = datetime.now(timezone.utc)
        self._commit("end", {"endTimestamp": self.end_timestamp})

    @staticmethod
    def _end_(
        writer: LogWriter,
        entity: Entity,
        id: str,
        data: Optional[Dict[str, Any]] = None,
    ):
        if data is None:
            data = {}
        data = {k: v for k, v in data.items() if v is not None}
        writer.commit(CommitLog(entity, id, "end", data))

    def data(self) -> Dict[str, Any]:
        data = {
            "name": self._name,
            "spanId": self.span_id,
            "tags": self.tags,
            "startTimestamp": self.start_timestamp,
            "endTimestamp": self.end_timestamp,
        }
        # removing none values
        data = {k: v for k, v in data.items() if v is not None}
        return data

    @staticmethod
    def _commit_(
        writer: LogWriter,
        entity: Entity,
        id: str,
        action: str,
        data: Optional[Dict[str, Any]] = None,
    ):
        # Removing all null values from data dict
        if data is not None:
            data = {k: v for k, v in data.items() if v is not None}
        writer.commit(CommitLog(entity, id, action, data))

    def _commit(self, action: str, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = self.data()
        # Removing all null values from data dict
        data = {k: v for k, v in data.items() if v is not None}
        self.writer.commit(CommitLog(self.entity, self._id, action, data))


class EventEmittingBaseContainer(BaseContainer):
    @staticmethod
    def _event_(
        writer: LogWriter,
        entity: Entity,
        entity_id: str,
        id: str,
        name: str,
        tags: Optional[Dict[str, str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        if metadata is not None:
            sanitized_metadata: dict[str, str] = _sanitize_metadata(metadata)
            BaseContainer._commit_(
                writer,
                entity,
                entity_id,
                "add-event",
                {
                    "id": id,
                    "name": name,
                    "timestamp": datetime.now(timezone.utc),
                    "tags": tags,
                    "metadata": sanitized_metadata,
                },
            )
        else:
            BaseContainer._commit_(
                writer,
                entity,
                entity_id,
                "add-event",
                {
                    "id": id,
                    "name": name,
                    "timestamp": datetime.now(timezone.utc),
                    "tags": tags,
                },
            )

    def event(
        self,
        id: str,
        name: str,
        tags: Optional[Dict[str, str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        if metadata is not None:
            sanitized_metadata: dict[str, str] = _sanitize_metadata(metadata)
            self._commit(
                "add-event",
                {
                    "id": id,
                    "name": name,
                    "timestamp": datetime.now(timezone.utc),
                    "tags": tags,
                    "metadata": sanitized_metadata,
                },
            )
        else:
            self._commit(
                "add-event",
                {
                    "id": id,
                    "name": name,
                    "timestamp": datetime.now(timezone.utc),
                    "tags": tags,
                },
            )



import atexit
import json
import os
import signal
import sys
import threading
import time
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, TypedDict, Union

from typing_extensions import deprecated

from .apis import MaximAPI
from .cache import MaximCache, MaximInMemoryCache
from .filter_objects import (
    IncomingQuery,
    QueryObject,
    find_all_matches,
    find_best_match,
    parse_incoming_query,
)
from .logger import Logger, LoggerConfig, LoggerConfigDict
from .models import (
    ChatCompletionMessage,
    Folder,
    FolderEncoder,
    Node,
    Prompt,
    PromptChain,
    PromptChainVersion,
    PromptChainVersionsAndRules,
    PromptNode,
    PromptResponse,
    PromptVersion,
    QueryRule,
    RuleGroupType,
    RuleType,
    Tool,
    VersionAndRulesWithPromptChainIdEncoder,
    VersionAndRulesWithPromptId,
    VersionAndRulesWithPromptIdEncoder,
    VersionsAndRules,
)
from .runnable import RunnablePrompt, RunnablePromptChain
from .scribe import scribe
from .test_runs import TestRunBuilder
from .version import current_version


class ConfigDict(TypedDict, total=False):
    api_key: Optional[str]
    base_url: Optional[str]
    cache: Optional[MaximCache]
    debug: Optional[bool]
    raise_exceptions: Optional[bool]
    prompt_management: Optional[bool]


@deprecated(
    "This class will be removed in a future version. Use {} which is TypedDict."
)
@dataclass
class Config:
    """
    A class representing the configuration for the Maxim SDK.

    Attributes:
        api_key (str): The API key for the Maxim instance.
        base_url (Optional[str], optional): The base URL for the Maxim instance. Defaults to "https://app.getmaxim.ai".
        cache (Optional[MaximCache], optional): The cache to use for the Maxim instance. Defaults to None.
        debug (Optional[bool], optional): Whether to enable debug logging. Defaults to False.
        raise_exceptions (Optional[bool], optional): Whether to raise exceptions during logging operations. Defaults to False.
        prompt_management (Optional[bool], optional): Whether to enable prompt management. Defaults to False.
    """

    api_key: Optional[str] = os.environ.get("MAXIM_API_KEY")
    base_url: Optional[str] = None
    cache: Optional[MaximCache] = None
    debug: Optional[bool] = False
    raise_exceptions: Optional[bool] = False
    prompt_management: Optional[bool] = False


EntityType = {"PROMPT": "PROMPT", "FOLDER": "FOLDER", "PROMPT_CHAIN": "PROMPT_CHAIN"}


def get_config_dict(config: Union[Config, ConfigDict]) -> dict[str, Any]:
    return (
        {
            "api_key": config.api_key,
            "base_url": config.base_url,
            "cache": config.cache or MaximInMemoryCache(),
            "debug": config.debug,
            "raise_exceptions": config.raise_exceptions,
            "prompt_management": config.prompt_management,
        }
        if isinstance(config, Config)
        else dict(config)
    )


class Maxim:
    _lock = threading.Lock()
    """
    A class representing the Maxim SDK.

    This class provides methods for interacting with the Maxim API.
    """

    def __init__(self, config: Union[Config, ConfigDict, None] = None):
        """
        Initializes a new instance of the Maxim class.

        Args:
            config (Config): The configuration for the Maxim instance.
        """
        # Thread-safe singleton pattern implementation
        with Maxim._lock:
            if hasattr(Maxim, "_instance"):
                raise RuntimeError(
                    "Maxim is already initialized. You can initialize it only once."
                )
            Maxim._instance = self
        self.has_cleaned_up = False
        atexit.register(self.cleanup)
        if threading.current_thread() is threading.main_thread():
            signal.signal(signal.SIGINT, self._signal_handler)
            signal.signal(signal.SIGTERM, self._signal_handler)
        self.ascii_logo = (
            f"\033[32m[MaximSDK] Initializing Maxim AI(v{current_version})\033[0m"
        )
        # Print the ASCII logo when initializing
        print(self.ascii_logo)
        if config is None:
            config = ConfigDict()
        final_config = get_config_dict(config)
        if final_config.get("api_key", None) is None:
            # Checking in the env variable
            api_key = os.environ.get("MAXIM_API_KEY", None)
            if api_key is None:
                raise ValueError(
                    "API key is required. Either set it in the environment variable MAXIM_API_KEY or pass it as a parameter to the Config object."
                )
            final_config["api_key"] = api_key

        self.base_url = final_config.get("base_url", None) or "https://app.getmaxim.ai"
        self.api_key = final_config.get("api_key", None)
        self.is_running = True
        self.raise_exceptions = final_config.get("raise_exceptions", False)
        self.maxim_api = MaximAPI(self.base_url, self.api_key)
        self.__is_debug = final_config.get("debug", False)
        self.__loggers: Dict[str, Logger] = {}
        self.prompt_management = final_config.get("prompt_management", False)
        if self.prompt_management:
            self.__cache = final_config.get("cache", MaximInMemoryCache())
            self.__sync_thread = threading.Thread(target=self.__sync_timer)
            self.__sync_thread.daemon = True
            self.__sync_thread.start()

    def enable_prompt_management(self) -> "Maxim":
        self.prompt_management = True
        if not self.__sync_thread.is_alive():
            self.__sync_thread = threading.Thread(target=self.__sync_timer)
            self.__sync_thread.daemon = True
            self.__sync_thread.start()
        return self

    def enable_exceptions(self, val: bool) -> "Maxim":
        self.raise_exceptions = val
        return self

    def __sync_timer(self):
        if not self.prompt_management:
            return
        while self.is_running:
            self.__sync_entities()
            time.sleep(60)

    def __sync_entities(self):
        scribe().debug("[MaximSDK] Syncing prompts and folders")
        if not self.prompt_management:
            return
        if not self.is_running:
            return
        self.__sync_prompts()
        self.__sync_prompt_chains()
        self.__syncFolders()
        scribe().debug("[MaximSDK] Syncing completed")

    def __sync_prompts(self):
        scribe().debug("[MaximSDK] Syncing prompts")
        try:
            prompts = self.maxim_api.get_prompts()
            scribe().debug(f"[MaximSDK] Found {len(prompts)} prompts")
            for prompt in prompts:
                try:
                    self.__cache.set(
                        self.__get_cache_key(EntityType["PROMPT"], prompt.promptId),
                        json.dumps(prompt, cls=VersionAndRulesWithPromptIdEncoder),
                    )
                except Exception as err:
                    scribe().error(f"[MaximSDK] Error while syncing prompts: {err}")
                    if self.raise_exceptions:
                        raise err
        except Exception as err:
            scribe().error(f"[MaximSDK] Error while syncing prompts: {err}")
            if self.raise_exceptions:
                raise err

    def __sync_prompt_chains(self):
        scribe().debug("[MaximSDK] Syncing prompt Chains")
        try:
            prompt_chains = self.maxim_api.get_prompt_chains()
            scribe().debug(f"[MaximSDK] Found {len(prompt_chains)} prompt chains")
            for prompt_chain in prompt_chains:
                try:
                    self.__cache.set(
                        self.__get_cache_key(
                            EntityType["PROMPT_CHAIN"], prompt_chain.promptChainId
                        ),
                        json.dumps(
                            prompt_chain, cls=VersionAndRulesWithPromptChainIdEncoder
                        ),
                    )
                except Exception as err:
                    scribe().error(
                        f"[MaximSDK] Error while syncing prompt chains: {err}"
                    )
                    if self.raise_exceptions:
                        raise err
        except Exception as err:
            scribe().error(f"[MaximSDK] Error while syncing prompts: {err}")
            if self.raise_exceptions:
                raise err

    def __syncFolders(self):
        scribe().debug("[MaximSDK] Syncing folders")
        try:
            folders = self.maxim_api.get_folders()
            scribe().debug(f"[MaximSDK] Found {len(folders)} folders")
            for folder in folders:
                try:
                    self.__cache.set(
                        self.__get_cache_key(EntityType["FOLDER"], folder.id),
                        json.dumps(folder, cls=FolderEncoder),
                    )
                except Exception as err:
                    scribe().error(f"[MaximSDK] Error while syncing folders: {err}")
                    if self.raise_exceptions:
                        raise err
        except Exception as err:
            scribe().error(f"[MaximSDK] Error while syncing folders: {err}")
            if self.raise_exceptions:
                raise err

    def __get_cache_key(self, entity: str, id: str) -> str:
        if entity == "PROMPT":
            return f"prompt:{id}"
        else:
            return f"folder:{id}"

    def __get_prompt_from_cache(self, key: str) -> Optional[VersionsAndRules]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        return VersionAndRulesWithPromptId.from_dict(json_data)

    def __get_all_prompts_from_cache(self) -> Optional[List[VersionsAndRules]]:
        keys = self.__cache.get_all_keys()
        if not keys:
            return None
        data = [self.__cache.get(key) for key in keys if key.startswith("prompt:")]
        prompt_list = []
        for d in data:
            if d is not None:
                json_data = json.loads(d)
                if "promptId" in json_data:
                    json_data.pop("promptId")
                prompt_list.append(VersionsAndRules.from_dict(json_data))
        return prompt_list

    def __get_prompt_chain_from_cache(
        self, key: str
    ) -> Optional[PromptChainVersionsAndRules]:
        data = self.__cache.get(key)
        if not data:
            return None
        parsed_data = json.loads(data)
        return PromptChainVersionsAndRules.from_dict(parsed_data)

    def __get_all_prompt_chains_from_cache(
        self,
    ) -> Optional[List[PromptChainVersionsAndRules]]:
        keys = self.__cache.get_all_keys()
        if not keys:
            return None
        data = [
            self.__cache.get(key) for key in keys if key.startswith("prompt_chain:")
        ]
        prompt_chain_list = []
        for d in data:
            if d is not None:
                json_data = json.loads(d)
                if "promptChainId" in json_data:
                    json_data.pop("promptChainId")
                prompt_chain_list.append(
                    PromptChainVersionsAndRules.from_dict(json_data)
                )
        return prompt_chain_list

    def __get_folder_from_cache(self, key: str) -> Optional[Folder]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        return Folder(**json_data)

    def __get_all_folders_from_cache(self) -> Optional[List[Folder]]:
        keys = self.__cache.get_all_keys()
        if not keys:
            return None
        data = [self.__cache.get(key) for key in keys if key.startswith("folder:")]
        return [Folder(**json.loads(d)) for d in data if d is not None]

    def __format_prompt(self, prompt_version: PromptVersion) -> Prompt:
        return Prompt(
            prompt_id=prompt_version.promptId,
            version_id=prompt_version.id,
            version=prompt_version.version,
            messages=prompt_version.config.messages if prompt_version.config else [],
            tags=prompt_version.config.tags if prompt_version.config else {},
            model_parameters=prompt_version.config.modelParameters
            if prompt_version.config
            else {},
            model=prompt_version.config.model if prompt_version.config else None,
            provider=prompt_version.config.provider if prompt_version.config else None,
        )

    def __format_prompt_chain(
        self, prompt_chain_version: PromptChainVersion
    ) -> PromptChain:
        prompt_nodes: List[Node] = []
        if prompt_chain_version.config is not None:
            for node in prompt_chain_version.config.nodes:
                if isinstance(node.content, PromptNode):
                    prompt_nodes.append(node)
        return PromptChain(
            prompt_chain_id=prompt_chain_version.promptChainId,
            version_id=prompt_chain_version.id,
            version=prompt_chain_version.version,
            nodes=prompt_nodes,
        )

    def __get_prompt_version_for_rule(
        self,
        prompt_version_and_rules: VersionsAndRules,
        rule: Optional[QueryRule] = None,
    ) -> Optional[Prompt]:
        if rule:
            incoming_query = IncomingQuery(
                query=rule.query, operator=rule.operator, exactMatch=rule.exact_match
            )
            objects = []
            for version_id, version_rules in prompt_version_and_rules.rules.items():
                for version_rule in version_rules:
                    if not version_rule.rules.query:
                        continue
                    if rule.scopes:
                        for key in rule.scopes.keys():
                            if key != "folder":
                                raise ValueError("Invalid scope added")
                    version = next(
                        (
                            v
                            for v in prompt_version_and_rules.versions
                            if v.id == version_id
                        ),
                        None,
                    )
                    if not version:
                        continue
                    query = version_rule.rules.query
                    if version.config and version.config.tags:
                        parsed_incoming_query = parse_incoming_query(
                            incoming_query.query
                        )
                        tags = version.config.tags
                        for key, value in tags.items():
                            if value is None:
                                continue
                            if parsed_incoming_query is None:
                                continue
                            incomingQueryFields = [
                                q.field for q in parsed_incoming_query
                            ]
                            if key in incomingQueryFields:
                                query.rules.append(
                                    RuleType(field=key, operator="=", value=value)
                                )
                    objects.append(QueryObject(query=query, id=version_id))

            deployed_version_object = find_best_match(objects, incoming_query)
            if deployed_version_object:
                deployed_version = next(
                    (
                        v
                        for v in prompt_version_and_rules.versions
                        if v.id == deployed_version_object.id
                    ),
                    None,
                )
                if deployed_version:
                    return self.__format_prompt(deployed_version)

        else:
            if prompt_version_and_rules.rules:
                for version_id, version_rules in prompt_version_and_rules.rules.items():
                    does_query_exist = any(
                        ruleElm.rules.query is not None for ruleElm in version_rules
                    )
                    if does_query_exist:
                        deployed_version = next(
                            (
                                v
                                for v in prompt_version_and_rules.versions
                                if v.id == version_id
                            ),
                            None,
                        )
                        if deployed_version:
                            return self.__format_prompt(deployed_version)
            else:
                return self.__format_prompt(prompt_version_and_rules.versions[0])

        if prompt_version_and_rules.fallbackVersion:
            return self.__format_prompt(prompt_version_and_rules.fallbackVersion)
        return None

    def __get_prompt_chain_version_for_rule(
        self,
        prompt_chain_version_and_rules: PromptChainVersionsAndRules,
        rule: Optional[QueryRule] = None,
    ) -> Optional[PromptChain]:
        if rule:
            incoming_query = IncomingQuery(
                query=rule.query, operator=rule.operator, exactMatch=rule.exact_match
            )
            objects = []
            for (
                version_id,
                version_rules,
            ) in prompt_chain_version_and_rules.rules.items():
                for version_rule in version_rules:
                    if not version_rule.rules.query:
                        continue
                    if rule.scopes:
                        for key in rule.scopes.keys():
                            if key != "folder":
                                raise ValueError("Invalid scope added")
                    version: Optional[PromptChainVersion] = None
                    for v in prompt_chain_version_and_rules.versions:
                        if v.id != version_id:
                            continue
                        version = v
                        break
                    if not version:
                        continue
                    if version_rule.rules.query is not None:
                        query = version_rule.rules.query
                        if isinstance(query, RuleGroupType):
                            objects.append(QueryObject(query=query, id=version_id))

            deployed_version_object = find_best_match(objects, incoming_query)
            if deployed_version_object:
                deployed_version = next(
                    (
                        v
                        for v in prompt_chain_version_and_rules.versions
                        if v.id == deployed_version_object.id
                    ),
                    None,
                )
                if deployed_version:
                    return self.__format_prompt_chain(deployed_version)

        else:
            if prompt_chain_version_and_rules.rules:
                for (
                    version_id,
                    version_rules,
                ) in prompt_chain_version_and_rules.rules.items():
                    does_query_exist = any(
                        ruleElm.rules.query is not None for ruleElm in version_rules
                    )
                    if does_query_exist:
                        deployed_version = next(
                            (
                                v
                                for v in prompt_chain_version_and_rules.versions
                                if v.id == version_id
                            ),
                            None,
                        )
                        if deployed_version:
                            return self.__format_prompt_chain(deployed_version)
            else:
                return self.__format_prompt_chain(
                    prompt_chain_version_and_rules.versions[0]
                )

        if prompt_chain_version_and_rules.fallbackVersion:
            return self.__format_prompt_chain(
                prompt_chain_version_and_rules.fallbackVersion
            )
        return None

    def __get_folders_for_rule(
        self, folders: List[Folder], rule: QueryRule
    ) -> List[Folder]:
        incoming_query = IncomingQuery(
            query=rule.query, operator=rule.operator, exactMatch=rule.exact_match
        )
        objects = []
        for folder in folders:
            query = RuleGroupType(rules=[], combinator="AND")
            if not folder.tags:
                continue
            parsed_incoming_query = parse_incoming_query(incoming_query.query)
            tags = folder.tags
            for key, value in tags.items():
                if key in [q.field for q in parsed_incoming_query]:
                    # if not isinstance(value,None):
                    query.rules.append(RuleType(field=key, operator="=", value=value))
            if not query.rules:
                continue
            objects.append(QueryObject(query=query, id=folder.id))

        folder_objects = find_all_matches(objects, incoming_query)
        ids = [fo.id for fo in folder_objects]
        return [f for f in folders if f.id in ids]

    def get_prompt(self, id: str, rule: QueryRule) -> Optional[RunnablePrompt]:
        """
        Retrieves a prompt based on the provided id and rule.

        Args:
            id (str): The id of the prompt.
            rule (QueryRule): The rule to match the prompt against.

        Returns:
            Optional[Prompt]: The prompt object if found, otherwise None.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. You can enable it by initializing Maxim with Config(...prompt_management=True)."
            )
        key = self.__get_cache_key("PROMPT", id)
        version_and_rules_with_prompt_id = self.__get_prompt_from_cache(key)
        if version_and_rules_with_prompt_id is None:
            version_and_rules_with_prompt_id = self.maxim_api.get_prompt(id)
            if len(version_and_rules_with_prompt_id.versions) == 0:
                return None
            self.__cache.set(
                id,
                json.dumps(
                    version_and_rules_with_prompt_id,
                    cls=VersionAndRulesWithPromptIdEncoder,
                ),
            )
        if not version_and_rules_with_prompt_id:
            return None
        prompt = self.__get_prompt_version_for_rule(
            version_and_rules_with_prompt_id, rule
        )
        if not prompt:
            return None
        # wrapping this prompt with Runnable
        return RunnablePrompt(prompt, self.maxim_api)

    def get_prompts(self, rule: QueryRule) -> List[RunnablePrompt]:
        """
        Retrieves all prompts that match the given rule.

        Args:
            rule (QueryRule): The rule to match the prompts against.

        Returns:
            List[Prompt]: A list of prompts that match the given rule.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        version_and_rules = self.__get_all_prompts_from_cache()
        prompts: list[RunnablePrompt] = []
        if version_and_rules is None or len(version_and_rules) == 0:
            self.__sync_entities()
            version_and_rules = self.__get_all_prompts_from_cache()
        if not version_and_rules:
            return []
        for v in version_and_rules:
            if rule.scopes:
                if "folder" not in rule.scopes.keys():
                    return []
                else:
                    if rule.scopes["folder"] != v.folderId:
                        continue
            prompt = self.__get_prompt_version_for_rule(v, rule)
            if prompt is not None:
                prompts.append(RunnablePrompt(prompt, self.maxim_api))
        if len(prompts) == 0:
            return []
        return prompts

    def get_prompt_chain(
        self, id: str, rule: QueryRule
    ) -> Optional[RunnablePromptChain]:
        """
        Retrieves a prompt chain based on the provided id and rule.

        Args:
            id (str): The id of the prompt chain.
            rule (QueryRule): The rule to match the prompt chain against.

        Returns:
            Optional[PromptChain]: The prompt chain object if found, otherwise None.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config by setting Config(...,prompt_management=True)."
            )
        key = self.__get_cache_key("PROMPT_CHAIN", id)
        version_and_rules = self.__get_prompt_chain_from_cache(key)
        if version_and_rules is None:
            version_and_rules = self.maxim_api.getPromptChain(id)
            if len(version_and_rules.versions) == 0:
                return None
            self.__cache.set(
                id,
                json.dumps(
                    version_and_rules, cls=VersionAndRulesWithPromptChainIdEncoder
                ),
            )
        if not version_and_rules:
            return None
        prompt_chain = self.__get_prompt_chain_version_for_rule(version_and_rules, rule)
        if not prompt_chain:
            return None
        # The reason we have created a separate class
        # called agents is for two reasons
        # 1. We are not able to import MaximAPI in models fir circular deps
        # 2. We want to set the foundations of agents for future
        return RunnablePromptChain(prompt_chain, maxim_api=self.maxim_api)

    def get_folder_by_id(self, id: str) -> Optional[Folder]:
        """
        Retrieves a folder based on the provided id.

        Args:
            id (str): The id of the folder.

        Returns:
            Optional[Folder]: The folder object if found, otherwise None.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        key = self.__get_cache_key("FOLDER", id)
        folder = self.__get_folder_from_cache(key)
        if folder is None:
            try:
                folder = self.maxim_api.get_folder(id)
                if not folder:
                    return None
                self.__cache.set(key, json.dumps(folder, cls=FolderEncoder))
            except Exception:
                return None
        return folder

    def get_folders(self, rule: QueryRule) -> List[Folder]:
        """
        Retrieves all folders that match the given rule.

        Args:
            rule (QueryRule): The rule to match the folders against.

        Returns:
            List[Folder]: A list of folders that match the given rule.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        folders = self.__get_all_folders_from_cache()
        if folders is None or len(folders) == 0:
            self.__sync_entities()
            folders = self.__get_all_folders_from_cache()
        if not folders:
            return []
        return self.__get_folders_for_rule(folders, rule)

    def logger(
        self, config: Optional[Union[LoggerConfig, LoggerConfigDict]] = None
    ) -> Logger:
        """
        Creates a logger based on the provided configuration.

        Args:
            config (LoggerConfig): The configuration for the logger.

        Returns:
            Logger: The logger object.
        """
        final_config: LoggerConfigDict = {}
        if config is not None:
            if isinstance(config, LoggerConfig):
                final_config = {
                    "id": config.id,
                    "auto_flush": config.auto_flush,
                    "flush_interval": config.flush_interval,
                }
            else:
                final_config = config

        if final_config.get("id", None) is None:
            repo_id = os.environ.get("MAXIM_LOG_REPO_ID")
            if repo_id is None:
                raise ValueError(
                    "Log repo id is required. Either set environment variable MAXIM_LOG_REPO_ID or pass it as a parameter in LoggerConfig."
                )
            final_config["id"] = repo_id

        repo_id = final_config.get("id", None)
        if repo_id is None:
            raise ValueError("Log repository id is required")
        if repo_id in self.__loggers:
            return self.__loggers[repo_id]
        logger = Logger(
            config=final_config,
            api_key=self.api_key,
            base_url=self.base_url,
            is_debug=self.__is_debug,
            raise_exceptions=self.raise_exceptions,
        )
        self.__loggers[repo_id] = logger

        return logger

    def _check_if_repo_exists(self, logger: Logger):
        """
        Checks if the log repository exists.
        """

        def check():
            try:
                exists = self.maxim_api.does_log_repository_exist(logger.repo_id)
                if not exists:
                    scribe().warning(f"[MaximSDK] Log repository not found: {logger.repo_id}. We will be dropping all logs.")
                    if self.raise_exceptions:
                        raise ValueError(f"Log repository not found: {logger.repo_id}")
                    return
                scribe().debug(f"[MaximSDK] Log repository found: {logger.repo_id}")
            except Exception as e:
                scribe().error(f"[MaximSDK] Failed to check repository existence: {str(e)}")
                if self.raise_exceptions:
                    raise            

        thread = threading.Thread(target=check)
        thread.daemon = True
        thread.start()

    def create_test_run(self, name: str, in_workspace_id: str) -> TestRunBuilder:
        """
        Creates a test run builder based on the provided name and workspace id.

        Args:
            name (str): The name of the test run.
            in_workspace_id (str): The workspace id to create the test run in.

        Returns:
            TestRunBuilder: The test run builder object.
        """
        return TestRunBuilder(
            name=name,
            workspace_id=in_workspace_id,
            api_key=self.api_key,
            base_url=self.base_url,
            evaluators=[],
        )

    def chat_completion(
        self,
        model: str,
        messages: List[ChatCompletionMessage],
        tools: Optional[List[Tool]] = None,
        **kwargs,
    ) -> Optional[PromptResponse]:
        """
        Performs a chat completion request using the specified model and messages.

        Args:
            model (str): The model name to use for completion. The expected format is "provider/model_name". Example "openai/gpt-3.5-turbo".
            messages (List[ChatCompletionMessage]): List of chat messages in the conversation
            tools (Optional[List[Tool]], optional): List of tools available to the model. Defaults to None.
            **kwargs: Additional model parameters to pass to the completion request

        Returns:
            Optional[PromptResponse]: The completion response if successful, None otherwise
        """
        return self.maxim_api.run_prompt(model, messages, tools, **kwargs)

    def _signal_handler(self, signum, frame):
        if self.has_cleaned_up:
            return
        self.has_cleaned_up = True
        self.cleanup()
        sys.exit(0)

    def cleanup(self):
        """
        Cleans up the Maxim sync thread.
        """
        if self.has_cleaned_up:
            return
        # We sleep to allow all pending writes in the queue to come in and then we can flush
        time.sleep(2)
        scribe().debug("[MaximSDK] Cleaning up Maxim sync thread")
        self.is_running = False
        for logger in self.__loggers.values():
            logger.cleanup(is_sync=True)
        scribe().debug("[MaximSDK] Cleanup done")



import ast
import re
import warnings
from typing import Any, Sequence, Union

EXCLUDE_TASK_ATTRS = {"agent": True}

EXCLUDE_AGENT_ATTRS = {
    "crew": True,
}


def is_primitive(obj: Any) -> bool:
    """Check if an object is a known primitive type."""
    return isinstance(obj, (int, float, str, bool, type(None)))


def dictify(
    obj: Any, maxdepth: int = 0, depth: int = 1, seen: Union[set[int], None] = None
) -> Any:
    """Recursively compute a dictionary representation of an object."""
    if seen is None:
        seen = set()

    if not is_primitive(obj):
        obj_id = id(obj)
        if obj_id in seen:
            # Avoid infinite recursion with circular references
            return stringify(obj)
        else:
            seen.add(obj_id)

    if maxdepth > 0 and depth > maxdepth:
        # TODO: If obj at this point is a simple type,
        #       maybe we should just return it rather than stringify
        return stringify(obj)

    if is_primitive(obj):
        return obj
    elif isinstance(obj, (list, tuple)):
        return [dictify(v, maxdepth, depth + 1, seen) for v in obj]
    elif isinstance(obj, dict):
        dict_result = {}
        for k, v in obj.items():
            if isinstance(k, str) and should_redact(k):
                dict_result[k] = REDACTED_VALUE
            else:
                dict_result[k] = dictify(v, maxdepth, depth + 1, seen)
        return dict_result

    if hasattr(obj, "to_dict"):
        try:
            as_dict = obj.to_dict()
            if isinstance(as_dict, dict):
                to_dict_result = {}
                for k, v in as_dict.items():
                    if isinstance(k, str) and should_redact(k):
                        to_dict_result[k] = REDACTED_VALUE
                    elif maxdepth == 0 or depth < maxdepth:
                        to_dict_result[k] = dictify(v, maxdepth, depth + 1, seen)
                    else:
                        to_dict_result[k] = stringify(v)
                return to_dict_result
        except Exception:
            # If to_dict fails, we can fall through to the generic object serialization
            # Or, if preferred, raise a more specific error or log it.
            # For now, let's make it fall through by not re-raising or returning here.
            pass  # Fall through to generic handling

    result: dict[Any, Any] = {}
    result["__class__"] = {
        "module": obj.__class__.__module__,
        "qualname": obj.__class__.__qualname__,
        "name": obj.__class__.__name__,
    }
    if isinstance(obj, Sequence) and not isinstance(obj, (str, bytes)):
        # Custom list-like object
        try:
            for i, item in enumerate(obj):
                result[i] = dictify(item, maxdepth, depth + 1, seen)
        except Exception:
            return stringify(obj)
    else:
        # Define Pydantic internal attributes to skip during generic iteration
        # as these are handled by model_dump (via to_dict) or are metadata.
        pydantic_internal_attrs = {
            "model_fields",
            "model_computed_fields",
            "model_config",
            "__slots__",
            "__pydantic_fields_set__",
            "__pydantic_extra__",
            "__pydantic_private__",
            # Adding common SQLAlchemy internal/deprecated attrs that might appear via dir()
            # though this doesn't fix the SADeprecationWarning origin, it avoids getattr on them here.
            "metadata",
            "_sa_instance_state",
            "registry",
            "full_returning",
        }
        for attr in dir(obj):
            if attr.startswith("_") or attr in pydantic_internal_attrs:
                continue
            if should_redact(attr):
                result[attr] = REDACTED_VALUE
                continue
            try:
                val = getattr(obj, attr)
                if callable(val):
                    continue
                if maxdepth == 0 or depth < maxdepth:
                    result[attr] = dictify(val, maxdepth, depth + 1, seen)
                else:
                    result[attr] = stringify(val)
            except Exception as e_attr:
                # Handle error for specific attribute instead of aborting the whole object
                result[attr] = stringify(
                    f"<Error serializing attribute {attr}: {e_attr}>"
                )
    return result


MAX_STR_LEN = 1000


def stringify(obj: Any, limit: int = MAX_STR_LEN) -> str:
    """This is a fallback for objects that we don't have a better way to serialize."""
    rep = None
    try:
        rep = repr(obj)
    except Exception:
        try:
            rep = str(obj)
        except Exception:
            rep = f"<{type(obj).__name__}: {id(obj)}>"
    if isinstance(rep, str):
        if len(rep) > limit:
            rep = rep[: limit - 3] + "..."
    return rep


def serialize_crewai_objects(obj: Any) -> Any:
    """Safely serialize CrewAI objects to prevent recursion."""
    # Return primitive types directly
    if obj is None or isinstance(obj, (str, int, float, bool)):
        return obj

    # Everything else is serialized as a dict
    if hasattr(obj, "__class__"):
        if obj.__class__.__name__ == "Agent":
            return serialize_crewai_agent(obj)
        elif obj.__class__.__name__ == "Task":
            return serialize_crewai_task(obj)
        else:
            return dictify(obj)


def serialize_crewai_agent(obj: Any) -> dict[str, Any]:
    # Ensure obj is a pydantic BaseModel
    if not hasattr(obj, "model_dump"):
        return {"type": "Agent", "error": "Not a valid Pydantic model"}

    result = {
        "type": "Agent",
    }

    # Core identity attributes. We want to surface these first.
    core_attributes = ["role", "goal", "backstory", "id"]
    for attr in core_attributes:
        if hasattr(obj, attr):
            value = getattr(obj, attr)
            if attr == "id" and value is not None:
                result[attr] = stringify(value)
            elif value is not None:
                result[attr] = value

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        attr_dict = obj.model_dump(
            exclude=EXCLUDE_AGENT_ATTRS,
            exclude_none=True,
        )

    for attr, value in attr_dict.items():
        result[attr] = stringify(value)

    return result


def serialize_crewai_task(obj: Any) -> dict[str, Any]:
    if not hasattr(obj, "model_dump"):
        return {"type": "Task", "error": "Not a valid Pydantic model"}

    result = {
        "type": "Task",
    }

    # Core identity attributes. We want to surface these first.
    core_attributes = ["name", "description", "expected_output", "id"]
    for attr in core_attributes:
        if hasattr(obj, attr):
            value = getattr(obj, attr)
            if attr == "id" and value is not None:
                result[attr] = stringify(value)
            elif value is not None:
                result[attr] = value

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        attr_dict = obj.model_dump(
            exclude=EXCLUDE_TASK_ATTRS,
            exclude_none=True,
        )

    for attr, value in attr_dict.items():
        if attr.startswith("_") or value == "":
            continue
        else:
            result[attr] = stringify(value)

    return result


def crewai_postprocess_inputs(inputs: dict[str, Any]) -> dict[str, Any]:
    """Process CrewAI inputs to prevent recursion."""
    return {k: serialize_crewai_objects(v) for k, v in inputs.items()}


def crew_kickoff_postprocess_inputs(inputs: dict[str, Any]) -> dict[str, Any]:
    """Postprocess the inputs to the Crew.kickoff method.

    The method has a self which should be an instance of `Crew` which is a pydantic model.
    The method also has an inputs which is a dict or list[dict] of arguments to pass to the `kickoff` method.
    """
    results = {}
    for k, v in inputs.items():
        if k == "self":
            if hasattr(v, "model_dump"):
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=UserWarning)
                    crew_dict = v.model_dump()
                    if isinstance(crew_dict, dict):
                        results["self"] = {
                            k2: v2
                            for k2, v2 in crew_dict.items()
                            if v2 is not None
                            and not (isinstance(v2, list) and len(v2) == 0)
                        }
                    else:
                        results["self"] = crew_dict
        if k == "inputs":
            results["inputs"] = dictify(v)

    return results


def get_task_display_name(inputs: dict) -> str:
    task_representation = inputs.get("self")
    name_value = None  # Variable to hold the raw value of name

    if isinstance(task_representation, dict):
        # If 'self' is a dictionary, get the 'name' value
        name_value = task_representation.get("name")
    elif task_representation is not None:
        # If 'self' is an object, get the 'name' attribute, defaulting to None if not found or if attribute itself is None
        name_value = getattr(task_representation, "name", None)

    # Process the name_value: if it's a string, clean it; otherwise, use an empty string
    if isinstance(name_value, str):
        name = name_value.replace("\\n", "").strip()
    else:
        name = ""  # Default to empty string if name_value is None or not a string

    if not name:  # Check the cleaned and stripped name
        return "crewai.Task"
    return name


def get_agent_display_name(inputs: dict) -> str:
    agent_representation = inputs.get("self")
    title = ""

    if isinstance(agent_representation, dict):
        # If 'self' is a dictionary (likely already serialized)
        title = agent_representation.get("role", "")
        if not title:
            title = agent_representation.get("name", "")
    elif agent_representation is not None:
        # If 'self' is not a dictionary but is not None,
        # assume it's an object (e.g., Agent instance) and try to access attributes.
        # Use getattr for safe access with a default empty string.
        title = getattr(agent_representation, "role", "")
        if not title:  # If role is not found or is empty, try name
            title = getattr(agent_representation, "name", "")

    if not title:  # If title is still empty after trying role and name
        return "crewai.Agent"

    # Ensure title is a string before calling string methods, then clean and strip
    return str(title).replace("\n", "").strip()


def extract_tool_name(text: str) -> str:
    match = re.search(r"\.([\w]+)\._run$", text)
    return match.group(1) if match else ""


def extract_tool_details(description: str) -> dict[str, Any]:
    try:
        name_match = re.search(r"Tool Name:\s*(.+)", description)
        tool_name = name_match.group(1).strip() if name_match else None
    except Exception:
        tool_name = None

    # Extract Tool Arguments as Python dict
    tool_args = {}
    try:
        args_match = re.search(
            r"Tool Arguments:\s*(\{.*?\})\s*Tool Description:", description, re.DOTALL
        )
        if args_match:
            tool_args = ast.literal_eval(args_match.group(1))
    except Exception:
        tool_args = None

    # Extract Tool Description
    try:
        desc_match = re.search(r"Tool Description:\s*(.+)", description)
        tool_description = desc_match.group(1).strip() if desc_match else None
    except Exception:
        tool_description = None

    return {
        "name": tool_name,
        "args": tool_args,
        "description": tool_description,
    }


REDACT_KEYS = (
    "api_key",
    "auth_headers",
    "authorization",
)
REDACTED_VALUE = "REDACTED"


def should_redact(key: str) -> bool:
    return key.lower() in REDACT_KEYS



"""
This module is a wrapper around the Portkey client that allows for easy integration with Maxim.

It instruments the Portkey client to log to Maxim.

It also provides a wrapper around the Portkey chat-completions client that allows for easy integration with Maxim.
"""
from typing import Optional, Union
from uuid import uuid4

from ...scribe import scribe
from ..logger import Generation, Logger, Trace
from ..openai.utils import OpenAIUtils

try:
    from portkey_ai import AsyncPortkey, Portkey  # type: ignore
except ImportError as e:
    raise ImportError(
        (
            "The 'portkey-ai' package is required for Portkey integration. "
            "Install it with `pip install portkey-ai` or `uv add portkey-ai`."
        )
    ) from e


class MaximPortkeyClient:
    """Maxim instrumenter for Portkey client that directly handles chat.completion method."""

    def __init__(self, client: Union[Portkey, AsyncPortkey], logger: Logger):
        self._client = client
        self._logger = logger

        # Handle both sync and async clients
        if isinstance(client, AsyncPortkey):
            self._chat = MaximAsyncPortkeyChat(client, logger)
        else:
            self._chat = MaximPortkeyChat(client, logger)

    @property
    def chat(self) -> Union["MaximPortkeyChat", "MaximAsyncPortkeyChat"]:
        return self._chat

    def __getattr__(self, name):
        """Delegate all other attributes to the underlying Portkey client."""
        return getattr(self._client, name)


class MaximPortkeyChat:
    """Maxim instrumenter for Portkey chat functionality."""

    def __init__(self, client: Portkey, logger: Logger):
        self._client = client
        self._logger = logger
        self._completions = MaximPortkeyChatCompletions(client, logger)

    @property
    def completions(self) -> "MaximPortkeyChatCompletions":
        return self._completions

    def __getattr__(self, name):
        """Delegate all other attributes to the underlying chat client."""
        return getattr(self._client.chat, name)


class MaximAsyncPortkeyChat:
    """Maxim instrumenter for async Portkey chat functionality."""

    def __init__(self, client: AsyncPortkey, logger: Logger):
        self._client = client
        self._logger = logger
        self._completions = MaximAsyncPortkeyChatCompletions(client, logger)

    @property
    def completions(self) -> "MaximAsyncPortkeyChatCompletions":
        return self._completions

    def __getattr__(self, name):
        """Delegate all other attributes to the underlying chat client."""
        return getattr(self._client.chat, name)


class MaximPortkeyChatCompletions:
    """Maxim instrumenter for Portkey chat completions."""

    def __init__(self, client: Portkey, logger: Logger):
        self._client = client
        self._logger = logger

    def create(self, *args, **kwargs):
        """Instrumented create method that logs to Maxim."""
        extra_headers = kwargs.get("extra_headers", None)
        trace_id = None
        generation_name = None

        if extra_headers is not None:
            trace_id = extra_headers.get("x-maxim-trace-id", None)
            generation_name = extra_headers.get("x-maxim-generation-name", None)
        is_local_trace = trace_id is None
        model = kwargs.get("model", None)
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None
        messages = kwargs.get("messages", None)

        try:
            trace = self._logger.trace({"id": final_trace_id})
            gen_config = {
                "id": str(uuid4()),
                "model": model,
                "provider": "portkey",
                "name": generation_name,
                "model_parameters": OpenAIUtils.get_model_params(**kwargs),
                "messages": OpenAIUtils.parse_message_param(messages),
            }
            generation = trace.generation(gen_config)
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][MaximPortkeyChatCompletions] Error in generating content: {str(e)}"
            )

        # Call the actual Portkey completion
        response = self._client.chat.completions.create(*args, **kwargs)

        try:
            if generation is not None:
                generation.result(OpenAIUtils.parse_completion(response))
            if is_local_trace and trace is not None:
                trace.set_output(response.choices[0].message.content or "")
                trace.end()
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][MaximPortkeyChatCompletions] Error in logging generation: {str(e)}"
            )

        return response

    def __getattr__(self, name):
        """Delegate all other attributes to the underlying completions client."""
        return getattr(self._client.chat.completions, name)


class MaximAsyncPortkeyChatCompletions:
    """Maxim instrumenter for async Portkey chat completions."""

    def __init__(self, client: AsyncPortkey, logger: Logger):
        self._client = client
        self._logger = logger

    async def create(self, *args, **kwargs):
        """Instrumented async create method that logs to Maxim."""
        extra_headers = kwargs.get("extra_headers", None)
        trace_id = None
        generation_name = None

        if extra_headers is not None:
            trace_id = extra_headers.get("x-maxim-trace-id", None)
            generation_name = extra_headers.get("x-maxim-generation-name", None)

        is_local_trace = trace_id is None
        model = kwargs.get("model", None)
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None
        messages = kwargs.get("messages", None)

        try:
            trace = self._logger.trace({"id": final_trace_id})
            gen_config = {
                "id": str(uuid4()),
                "model": model,
                "provider": "portkey",
                "name": generation_name,
                "model_parameters": OpenAIUtils.get_model_params(**kwargs),
                "messages": OpenAIUtils.parse_message_param(messages),
            }
            generation = trace.generation(gen_config)
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][MaximAsyncPortkeyChatCompletions] Error in generating content: {str(e)}"
            )

        # Call the actual async Portkey completion
        response = await self._client.chat.completions.create(*args, **kwargs)

        try:
            if generation is not None:
                generation.result(OpenAIUtils.parse_completion(response))
            if is_local_trace and trace is not None:
                trace.set_output(response.choices[0].message.content or "")
                trace.end()
        except Exception as e:
            scribe().warning(
                f"[MaximSDK][MaximAsyncPortkeyChatCompletions] Error in logging generation: {str(e)}"
            )

        return response

    def __getattr__(self, name):
        """Delegate all other attributes to the underlying completions client."""
        return getattr(self._client.chat.completions, name)



from openai import AsyncOpenAI, OpenAI

from ..logger import (
    Logger,
)
from .async_client import MaximOpenAIAsyncClient
from .chat import MaximOpenAIChat


class MaximOpenAIClient:
    def __init__(self, client: OpenAI, logger: Logger):
        self._client = client
        self._logger = logger
        self._aio = MaximOpenAIAsyncClient(AsyncOpenAI(api_key=client.api_key), logger)

    @property
    def chat(self) -> MaximOpenAIChat:
        return MaximOpenAIChat(self._client, self._logger)

    @property
    def aio(self) -> MaximOpenAIAsyncClient:
        return self._aio
